import torch
import torch.nn as nn
from optrade.torch.models.utils.weight_init import xavier_init
from optrade.torch.models.utils.revin import RevIN
from optrade.torch.models.utils.utils import Reshape

from typing import Optional

# Taken from: https://github.com/thuml/Time-Series-Library/blob/main/models/TSMixer.py


class ResBlock(nn.Module):
    def __init__(
        self,
        seq_len,
        d_model,
        dropout,
        num_channels,
    ):
        super(ResBlock, self).__init__()

        self.temporal = nn.Sequential(
            nn.Linear(seq_len, d_model),
            nn.GELU(),
            nn.Linear(d_model, seq_len),
            nn.Dropout(dropout),
        )

        self.channel = nn.Sequential(
            nn.Linear(num_channels, d_model),
            nn.GELU(),
            nn.Linear(d_model, num_channels),
            nn.Dropout(dropout),
        )

    def forward(self, x):
        """
        Args:
            x (torch.Tensor): Shape (batch_size, seq_len, num_vars)

        Returns:
            torch.Tensor: Shape (batch_size, seq_len, num_vars)
        """
        x = x + self.temporal(x.transpose(1, 2)).transpose(1, 2)
        x = x + self.channel(x)

        return x


class TSMixer(nn.Module):
    def __init__(
        self,
        seq_len: int,
        pred_len: int,
        num_enc_layers: int,
        d_model: int,
        num_channels: int,
        dropout: float = 0.0,
        revin: bool = True,
        revin_affine: bool = True,
        revout: bool = False,
        eps_revin: float = 1e-5,
        return_head: bool = True,
        target_channels: Optional[list] = None,
        channel_independent: bool = False,
    ) -> None:
        super(TSMixer, self).__init__()

        # Parameters
        self.num_enc_layers = num_enc_layers
        self.eps_revin = eps_revin
        self.revin_affine = revin_affine
        self.revout = revout
        self.target_channels = target_channels
        self.num_channels = num_channels
        self.pred_len = pred_len
        self.seq_len = seq_len
        self.return_head = return_head

        # Layers
        self.backbone = nn.ModuleList(
            [
                ResBlock(seq_len, d_model, dropout, num_channels)
                for _ in range(num_enc_layers)
            ]
        )

        if channel_independent:
            self.head = nn.Linear(seq_len, pred_len)
        else:
            if target_channels is not None:
                num_output_channels = len(target_channels)
            else:
                num_output_channels = num_channels

            self.head = nn.Sequential(
                Reshape(-1, num_output_channels * seq_len),
                nn.Linear(
                    num_output_channels * seq_len, num_output_channels * pred_len
                ),
                Reshape(-1, num_output_channels, pred_len),
            )

        # Initialize layers
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        self.apply(xavier_init)

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def forward(self, x):

        # RevIN
        if self._revin:
            x = self.revin(x, mode="norm")

        x = x.permute(
            0, 2, 1
        )  # (batch_size, num_channels, seq_len) => (batch_size, seq_len, num_channels)
        for i in range(self.num_enc_layers):
            x = self.backbone[i](x)

        if self.target_channels is not None:
            x = x[:, :, self.target_channels].transpose(
                1, 2
            )  # (batch_size, len(target_channels), seq_len)
        else:
            x = x.transpose(1, 2)  # (batch_size, num_channels, seq_len)

        if self.return_head:
            out = self.head(
                x
            )  # (batch_size, seq_len, len(target_channels)) => (batch_size, len(target_channels), pred_len)
        else:
            out = x

        # RevOUT
        if self.revout:
            out = self.revin(out, mode="denorm")

        return out


# Test
if __name__ == "__main__":
    batch_size = 32
    seq_len = 512
    pred_len = 96
    num_channels = 7
    num_enc_layers = 3
    d_model = 16
    dropout = 0.1
    revin = True
    revin_affine = True
    revout = True
    eps_revin = 1e-5

    x = torch.rand(batch_size, num_channels, seq_len)

    model = TSMixer(
        seq_len=seq_len,
        pred_len=pred_len,
        num_enc_layers=num_enc_layers,
        d_model=d_model,
        dropout=dropout,
        num_channels=num_channels,
        revin=revin,
        revin_affine=revin_affine,
        revout=revout,
        eps_revin=eps_revin,
        target_channels=[4, 5],
        channel_independent=True,
    )

    y = model(x)
    print(f"x: {x.shape} => y: {y.shape}")



import torch
import torch.nn as nn


def get_num_examples(data_loader):
    total_samples = len(data_loader.dataset)
    if not data_loader.drop_last:
        return total_samples
    else:
        batch_size = data_loader.batch_size
        return (total_samples // batch_size) * batch_size


class Transpose(nn.Module):
    def __init__(self, dim1, dim2):
        super().__init__()
        self.dim1, self.dim2 = dim1, dim2

    def forward(self, x):
        return x.transpose(self.dim1, self.dim2)


class Unsqueeze(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, x):
        return x.unsqueeze(self.dim)


class Reshape(nn.Module):
    def __init__(self, *args):
        super(Reshape, self).__init__()
        self.shape = args

    def forward(self, x):
        return x.reshape(*self.shape)


class Norm(nn.Module):
    def __init__(self, norm_mode, num_channels, seq_len, d_model):
        super().__init__()
        self.norm_mode = norm_mode
        self.num_channels = num_channels
        self.seq_len = seq_len
        self.d_model = d_model

        if norm_mode == "batch1d":
            self.norm = nn.Sequential(
                Transpose(1, 2), nn.BatchNorm1d(d_model), Transpose(1, 2)
            )
        elif norm_mode == "batch2d":
            self.norm = nn.BatchNorm2d(num_channels)
        elif norm_mode == "layer":
            self.norm = nn.LayerNorm(d_model)
        else:
            raise ValueError(
                "Please select a valid normalization mode: 'batch1d', 'batch2d', or 'layer'."
            )

    def forward(self, x):
        if self.norm_mode == "batch2d":
            batch_size = x.shape[0]
            x = x.view(batch_size, self.num_channels, self.seq_len, self.d_model)
            x = self.norm(x)
            x = x.view(batch_size * self.num_channels, self.seq_len, self.d_model)
            return x
        else:
            return self.norm(x)



import os
import json
import shutil
import neptune
from pydantic import BaseModel
from rich.console import Console
from datetime import datetime
from typing import Any, Dict

# Import yaml
import yaml

from optrade.config.config import Global


def offline_to_neptune():
    console = Console()
    log_dir = "logs"
    os.makedirs(log_dir, exist_ok=True)

    if not os.listdir(log_dir):
        console.log("[bold red]No logs to upload[/bold red]")
        return

    for exp in os.listdir(log_dir):
        exp_dir = os.path.join(log_dir, exp)
        if not os.path.isdir(exp_dir):
            console.log(f"[yellow]Skipping {exp_dir} as it's not a directory[/yellow]")
            continue

        exp_contents = os.listdir(exp_dir)
        if not exp_contents:
            console.log(f"[yellow]No contents found in {exp_dir}. Skipping...[/yellow]")
            continue

        processed_successfully = False
        if any(os.path.isfile(os.path.join(exp_dir, item)) for item in exp_contents):
            processed_successfully = process_experiment(exp_dir, console)
        else:
            processed_successfully = True
            for run in exp_contents:
                run_dir = os.path.join(exp_dir, run)
                if os.path.isdir(run_dir):
                    if not process_experiment(run_dir, console):
                        processed_successfully = False

        if processed_successfully:
            move_to_old_logs(exp_dir, console)
        else:
            console.log(
                f"[yellow]Skipping move of {exp} due to processing errors[/yellow]"
            )


def process_experiment(directory, console):
    try:
        json_file = find_json_file(directory, console)
        if not json_file:
            return False

        logs = read_json_file(directory, json_file, console)
        if not logs:
            return False

        neptune_logger = initialize_neptune_logger(logs, console)
        if not neptune_logger:
            return False

        upload_model_weights(directory, logs, neptune_logger, console)
        log_metrics_and_parameters(logs, neptune_logger, console)

        neptune_logger.stop()
        console.log(f"[green]Successfully processed and uploaded {directory}[/green]")
        return True

    except Exception as e:
        console.log(
            f"[bold red]Unexpected error processing {directory}: {str(e)}[/bold red]"
        )
        return False


def find_json_file(directory, console):
    try:
        return next(f for f in os.listdir(directory) if f.endswith(".json"))
    except StopIteration:
        console.log(f"[blue]No JSON file found in {directory}. Skipping...[/blue]")
        return None


def read_json_file(directory, json_file, console):
    try:
        with open(os.path.join(directory, json_file), "r") as f:
            return json.load(f)
    except json.JSONDecodeError:
        console.log(
            f"[bold red]Error: Unable to parse JSON file in {directory}[/bold red]"
        )
    except IOError:
        console.log(
            f"[bold red]Error: Unable to read JSON file in {directory}[/bold red]"
        )
    return None


def initialize_neptune_logger(logs, console):
    try:
        api_token = os.environ.get("NEPTUNE_API_TOKEN", "")
        return neptune.init_run(
            project=logs["parameters/exp/project_name"],
            api_token=api_token,
            custom_run_id=datetime.now().strftime("%Y-%m-%d_%H-%M-%S"),
        )
    except Exception as e:
        console.log(f"[bold red]Error initializing Neptune logger: {str(e)}[/bold red]")
        return None


def upload_model_weights(directory, logs, neptune_logger, console):
    try:
        model_weights = next(f for f in os.listdir(directory) if f.endswith(".pth"))
        model_id = logs["parameters/exp/model_id"]
        neptune_logger[f"model_checkpoints/{model_id}"].upload(
            os.path.join(directory, model_weights)
        )
    except StopIteration:
        console.log(
            f"[blue]No .pth file found in {directory}. Skipping model weight upload...[/blue]"
        )
    except Exception as e:
        console.log(f"[bold red]Error uploading model weights: {str(e)}[/bold red]")


def log_metrics_and_parameters(logs, neptune_logger, console):
    for key, value in logs.items():
        try:
            if isinstance(value, list):
                for v in value:
                    neptune_logger[key].append(v)
            else:
                neptune_logger[key] = value
        except Exception as e:
            console.log(f"[bold red]Error logging {key}: {str(e)}[/bold red]")


def move_to_old_logs(source_dir, console):
    destination_dir = "old_logs"
    destination_path = os.path.join(destination_dir, os.path.basename(source_dir))
    os.makedirs(destination_dir, exist_ok=True)
    console.log(
        f"Attempting to move run folder from {source_dir} to {destination_path}"
    )
    try:
        shutil.move(source_dir, destination_path)
        console.log(
            f"[green]Successfully moved {source_dir} to {destination_dir}[/green]"
        )
    except Exception as e:
        console.log(f"[bold red]Error moving {source_dir}: {str(e)}[/bold red]")

    if os.path.exists(destination_path):
        if not os.path.exists(source_dir):
            console.log(f"[green]Verified: {source_dir} was moved successfully[/green]")
        else:
            console.log(
                f"[yellow]Warning: {source_dir} exists in both source and destination[/yellow]"
            )
    else:
        console.log(
            f"[bold red]Error: {source_dir} was not moved to the destination[/bold red]"
        )


def epoch_logger(args, logger, key, value):
    if args.exp.neptune:
        logger[key].append(value)
    else:
        if key not in logger:
            logger[key] = []
        logger[key].append(value)


def log_pydantic(logger, obj, key="parameters"):
    def log_model(model, current_key):
        for attr_name, attr_value in model.__dict__.items():
            if isinstance(attr_value, BaseModel):
                log_model(attr_value, f"{current_key}/{attr_name}")
            else:
                converted_value = convert_pydantic_types(attr_value)
                logger[f"{current_key}/{attr_name}"] = converted_value

    if isinstance(obj, BaseModel):
        log_model(obj, key)
    elif hasattr(obj, "__dict__"):
        for attr_name, attr_value in obj.__dict__.items():
            if isinstance(attr_value, BaseModel):
                log_model(attr_value, f"{key}/{attr_name}")
            else:
                converted_value = convert_pydantic_types(attr_value)
                logger[f"{key}/{attr_name}"] = converted_value
    else:
        print(f"Unsupported type for Neptune logging: {type(obj)}")


def convert_pydantic_types(value):
    if isinstance(value, list):
        return ",".join(str(v) for v in value)
    elif isinstance(value, tuple):
        return str(value)
    elif isinstance(value, dict):
        return {k: convert_pydantic_types(v) for k, v in value.items()}
    else:
        return value


# Converting PyDantic -> YAML for Neptune upload
def represent_none(self, _):
    return self.represent_scalar("tag:yaml.org,2002:null", "")


yaml.SafeDumper.add_representer(type(None), represent_none)


class CleanDumper(yaml.SafeDumper):
    def represent_tuple(self, data):
        return self.represent_list(list(data))


CleanDumper.add_representer(tuple, CleanDumper.represent_tuple)


def global_to_yaml(global_instance: Global, file_path: str) -> str:
    def convert_to_dict(obj: Any) -> Dict:
        if isinstance(obj, BaseModel):
            return {
                k: convert_to_dict(v) for k, v in obj.dict().items() if v is not None
            }
        elif isinstance(obj, (list, tuple)):
            return [convert_to_dict(item) for item in obj]
        elif isinstance(obj, dict):
            return {k: convert_to_dict(v) for k, v in obj.items() if v is not None}
        else:
            return obj

    global_dict = convert_to_dict(global_instance)
    yaml_string = yaml.dump(global_dict, Dumper=CleanDumper, default_flow_style=False)

    if file_path:
        with open(file_path, "w") as file:
            file.write(yaml_string)

    return yaml_string


if __name__ == "__main__":
    offline_to_neptune()



import torch
import torch.nn as nn


class moving_avg(nn.Module):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size, stride):
        super(moving_avg, self).__init__()
        self.kernel_size = kernel_size
        self.avg = nn.AvgPool1d(kernel_size=kernel_size, stride=stride, padding=0)

    def forward(self, x):
        # padding on the both ends of time series
        front = x[:, 0:1, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        end = x[:, -1:, :].repeat(1, (self.kernel_size - 1) // 2, 1)
        x = torch.cat([front, x, end], dim=1)
        x = self.avg(x.permute(0, 2, 1))
        x = x.permute(0, 2, 1)
        return x


class series_decomp(nn.Module):
    """
    Series decomposition block
    """

    def __init__(self, kernel_size):
        super(series_decomp, self).__init__()
        self.moving_avg = moving_avg(kernel_size, stride=1)

    def forward(self, x):
        moving_mean = self.moving_avg(x)
        res = x - moving_mean
        return res, moving_mean



import torch
import torch.nn as nn


class Patcher(nn.Module):
    """
    Splits the input time series into patches.
    """

    def __init__(self, patch_dim: int = 16, stride: int = 8):
        super(Patcher, self).__init__()
        self.patch_dim = patch_dim
        self.stride = stride

    def forward(self, x):
        """
        Args:
            x: input tensor of shape (B, M, L). B: batch_size, M: channels, L: sequence_length.
        Returns:
            patches: tensor of shape (B, M, N, P). N: number of patches, P: patch_dim.
            patches_combined: tensor of shape (B * M, N, P). N: number of patches, P: patch_dim. This is more efficient
                              to input into the Transformer encoder, as we are applying it to channels independently, thus,
                              we can combine the batch and channel dimensions and then reshape it afterwards.
        """
        B, M, L = x.shape

        # Number of patches.
        N = int((L - self.patch_dim) / self.stride) + 2

        # Pad the time series with the last value on each channel repeated S times
        last_column = x[:, :, -1:]  # index
        padding = last_column.repeat(1, 1, self.stride)
        x = torch.cat((x, padding), dim=2)

        # Extract patches
        patches = x.unfold(
            dimension=2, size=self.patch_dim, step=self.stride
        )  # Unfold the input tensor to extract patches.
        patches = patches.contiguous().view(
            B, M, N, self.patch_dim
        )  # Reshape the tensor to (B, M, N, P).
        patches_combined = patches.view(
            B * M, N, self.patch_dim
        )  # Reshape the tensor to (B * M, N, P).

        return patches


class VerticalPatcher(nn.Module):
    """
    Splits the input time series into patches, vertically stacking channels within each patch.
    """

    def __init__(self, patch_dim: int = 16, stride: int = 8):
        super(VerticalPatcher, self).__init__()
        self.patch_dim = patch_dim
        self.stride = stride

    def forward(self, x):
        """
        Args:
            x: input tensor of shape (B, M, L). B: batch_size, M: channels, L: sequence_length.
        Returns:
            patches: tensor of shape (B, N, M*P). N: number of patches, M*P: channels * patch_dim
        """
        B, M, L = x.shape

        # Number of patches
        N = ((L - self.patch_dim) // self.stride) + 1

        # Pad if needed
        if (L - self.patch_dim) % self.stride != 0:
            pad_size = self.stride - ((L - self.patch_dim) % self.stride)
            last_values = x[:, :, -1:]
            padding = last_values.repeat(1, 1, pad_size)
            x = torch.cat((x, padding), dim=2)
            N = ((x.shape[2] - self.patch_dim) // self.stride) + 1

        # Initialize output tensor
        patches = torch.zeros(B, N, M * self.patch_dim, device=x.device)

        # Create patches
        for i in range(N):
            start_idx = i * self.stride
            end_idx = start_idx + self.patch_dim

            # Extract and stack patches from each channel
            for c in range(M):
                patch_start = c * self.patch_dim
                patch_end = (c + 1) * self.patch_dim
                patches[:, i, patch_start:patch_end] = x[:, c, start_idx:end_idx]

        return patches



import torch
import torch.nn as nn
from optrade.torch.models.utils.utils import Norm


class SupervisedHead(nn.Module):
    def __init__(self, linear_dim, pred_len, dropout=0.0):
        super().__init__()
        """
        Flattens and applies a linear layer to each channel independently to form a prediction.
        Args:
            num_channels (int): The number of channels in the input.
            linear_dim (int): The dimension of the linear layer, should be num_patches * d_model.
            pred_len (int): The length of the forecast window.
            dropout (float): The dropout value.
        """

        self.dropout = nn.Dropout(dropout)
        self.linear = nn.Linear(linear_dim, pred_len)

    def forward(self, x) -> torch.Tensor:
        """
        Applies a linear layer to each channel independently to form a prediction, optional dropout.
        Args:
            x (torch.Tensor): The input of shape (batch_size, num_channels, num_patches, d_model)
        Returns:
            x (torch.Tensor): The output of shape (batch_size, num_channels, pred_len).
        """
        x = self.linear(x)
        x = self.dropout(x)
        return x


class EncoderBlock(nn.Module):
    """
    Args:
        d_model: The embedding dimension.
        num_heads: The number of heads in the multi-head attention models.
        dropout: The dropout value.
        batch_first: If True, then the input and output tensors are provided as (batch, seq, feature).
        norm: The type of normalization to use. Either "batch1d", "batch2d", or "layer".
    """

    def __init__(
        self,
        d_model,
        d_ff,
        num_heads,
        num_channels,
        num_patches,
        attn_dropout=0.0,
        ff_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
    ):
        super(EncoderBlock, self).__init__()

        # Layers
        self.attn = _MultiheadAttention(
            num_heads=num_heads,
            d_model=d_model,
            dropout=attn_dropout,
            batch_first=batch_first,
        )
        self.ff = nn.Sequential(
            nn.Linear(d_model, d_ff),
            nn.GELU(),
            nn.Dropout(ff_dropout),
            nn.Linear(d_ff, d_model),
        )

        # Normalization
        self.norm = Norm(norm_mode, num_channels, num_patches, d_model)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: Input tensor of shape (batch_size, num_patches, d_model).
        Returns:
            fc_out: Output of the transformer block, a tensor of shape (batch_size, num_patches, d_model).
        """

        # Multihead Attention -> Add & Norm
        attn_out, _ = self.attn(x, x, x)
        attn_norm = self.norm(
            attn_out + x
        )  # Treat the input as the query, key and value for MHA.

        # Feedforward layer -> Add & Norm
        fc_out = self.ff(attn_norm)
        fc_norm = self.norm(fc_out + attn_out)

        return fc_norm


class _MultiheadAttention(nn.Module):
    """
    Multihead Attention mechanism from the Vanilla Transformer, with some preset parameters for the PatchTST model.
    """

    def __init__(self, num_heads: int, d_model: int, dropout=0.0, batch_first=True):
        super(_MultiheadAttention, self).__init__()

        # Layers
        self.attn = nn.MultiheadAttention(
            embed_dim=d_model,
            num_heads=num_heads,
            dropout=dropout,
            batch_first=batch_first,
        )

    def forward(
        self, Q: torch.Tensor, K: torch.Tensor, V: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            Q: Query embedding of shape: (batch_size, num_patches, d_model).
            K: Key embedding of shape (batch_size, num_patches, d_model).
            V: Value embedding of shape (batch_size, num_patches, d_model).
            batch_size: The batch size.
            num_patches: The sequence length.
            d_model: The embedding dimension.
        Returns:
            x: The output of the attention layer of shape (batch_size, num_patches, d_model).
        """
        return self.attn(query=Q, key=K, value=V, need_weights=False)



import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, List, Dict, Literal
import os
from datetime import datetime

# Custom modules
from optrade.data.thetadata import load_all_data


def set_theme(theme: Literal["light", "dark"] = "light"):
    """Set the plotting theme parameters."""
    base_params = {
        "text.usetex": True,
        "font.family": "serif",
        "font.serif": ["Computer Modern Roman"],
        "axes.titlesize": 20,
        "axes.labelsize": 18,  # Increased from 16
        "xtick.labelsize": 18,
        "ytick.labelsize": 16,
        "axes.spines.top": True,
        "axes.spines.right": True,
        "axes.linewidth": 1.5,
        "axes.grid": False,
    }

    if theme == "dark":
        theme_params = {
            **base_params,
            "figure.facecolor": "black",
            "axes.facecolor": "black",
            "text.color": "white",
            "axes.labelcolor": "white",
            "xtick.color": "white",
            "ytick.color": "white",
        }
        plot_colors = {
            "face": "black",
            "text": "white",
            "line": "white",
            "grid": "white",
            "spine": "white",
        }
    else:  # light theme
        theme_params = {
            **base_params,
            "figure.facecolor": "white",
            "axes.facecolor": "white",
            "text.color": "black",
            "axes.labelcolor": "black",
            "xtick.color": "black",
            "ytick.color": "black",
        }
        plot_colors = {
            "face": "white",
            "text": "black",
            "line": "black",
            "grid": "gray",
            "spine": "black",
        }

    plt.rcParams.update(theme_params)
    return plot_colors


# def analyze_time_series(
#     data_list: List[np.ndarray],
#     info_list: List[Dict],
#     dates: np.ndarray,
#     dpi: int = 600,
#     output_format: str = "eps",
#     theme: Literal['light', 'dark'] = 'light',
#     save_dir: Optional[str] = None,
#     volume: bool = False,
#     volume_list: Optional[List[np.ndarray]] = None,
# ) -> None:
#     """
#     Analyze and visualize time series data with improved LaTeX formatting and larger font sizes.

#     Args:
#         data_list (List[np.ndarray]): List of time series datasets.
#         info_list (List[Dict]): List of dictionaries containing information for each time series datasets.
#         dates_list (List[np.ndarray]): List of date strings numpy arrays for each dataset. Default is None.
#         dpi (int): DPI for saving plots. Default is 600.
#         output_format (str): Output format for saving plots. Default is "eps".
#         theme (str): Plot theme, either 'light' or 'dark'. Default is 'light'.
#     """
#     # Set theme and get colors
#     colors = set_theme(theme)

#     # Create output directory
#     output_dir = os.path.abspath(os.path.join(os.getcwd(), '..', '..', '..', 'optrade', 'figures', 'general'))
#     os.makedirs(output_dir, exist_ok=True)


#     for i in range(len(data_list)):
#         # Calculate global min and max for amplitude axis
#         global_min = np.min(data_list[i])
#         global_max = np.max(data_list[i])

#         # Add some padding to the limits
#         y_range = global_max - global_min
#         global_min -= y_range * 0.05
#         global_max += y_range * 0.05

#         # Individual channel analysis
#         data = data_list[i].squeeze()

#         # Get title and y-axis label from info_list
#         title = info_list[i]["title"]
#         y_axis = info_list[i]["y-axis"]

#         # Create figure
#         fig = plt.figure(figsize=(12, 8), facecolor=colors['face'])
#         ax = fig.add_subplot(111)

#         # Original plotting code
#         x_indices = list(range(len(dates)))
#         ax.plot(x_indices, data, color=colors['line'], linewidth=1.2)

#         # Title and labels
#         ax.set_title(fr"\textbf{{{title}}}", fontsize=32, pad=20, color=colors['text'])
#         ax.set_ylabel(r"\textrm{" + y_axis + r"}", fontsize=26, color='black')
#         ax.set_ylim(global_min, global_max)

#         # Set tick size for y-axis
#         plt.yticks(fontsize=26)

#         # # Set up ticks
#         # num_ticks = 5
#         # step = max(len(x_indices) // num_ticks, 1)
#         # selected_indices = x_indices[::step]
#         # selected_dates = [dates[i] for i in selected_indices]
#         # ax.set_xticks(selected_indices)
#         # ax.set_xticklabels(selected_dates, fontsize=26, color='black')

#         # # Common styling for both cases
#         # # Center align the date labels
#         # for tick in ax.get_xticklabels():
#         #     tick.set_ha('center')

#         num_ticks = 5
#         step = max(len(x_indices) // num_ticks, 1)
#         selected_indices = x_indices[::step]
#         selected_dates = [dates[i] for i in selected_indices]

#         # Create axis with adjusted parameters
#         ax.set_xticks(selected_indices)
#         ax.set_xticklabels(selected_dates, fontsize=26, color='black', rotation=0)

#         # Enhance bottom spine visibility
#         ax.spines['bottom'].set_color('black')
#         ax.spines['bottom'].set_linewidth(1.5)

#         # Adjust margins to prevent cutoff
#         plt.subplots_adjust(bottom=0.2, right=0.95)

#         # Properly align tick labels
#         for tick in ax.get_xticklabels():
#             tick.set_ha('center')
#             tick.set_va('top')


#         # Style the spines
#         for spine in ax.spines.values():
#             spine.set_color(colors['spine'])
#             spine.set_linewidth(1.5)

#         # Add subtle grid
#         ax.grid(True, axis='y', linestyle='--', alpha=0.2, color=colors['grid'])

#         # Adjust layout
#         plt.subplots_adjust(bottom=0.2)

#         # Save with tight layout
#         figure_name = info_list[i]["file_codename"]
#         plt.savefig(
#             os.path.join(output_dir, f"{figure_name}.{output_format}"),
#             dpi=dpi,
#             bbox_inches='tight',
#             facecolor=colors['face'],
#             edgecolor='none'
#         )
#         plt.close()


def analyze_time_series(
    data_list: List[np.ndarray],
    info_list: List[Dict],
    dates: np.ndarray,
    dpi: int = 600,
    output_format: str = "eps",
    theme: Literal["light", "dark"] = "light",
    save_dir: Optional[str] = None,
    volume: bool = True,
    volume_list: Optional[List[np.ndarray]] = None,
) -> None:
    """
    Analyze and visualize time series data with normalized volume subplot.
    """
    from sklearn.preprocessing import StandardScaler

    # Set theme and get colors
    colors = set_theme(theme)

    # Create output directory
    output_dir = os.path.abspath(
        os.path.join(os.getcwd(), "..", "..", "..", "optrade", "figures", "general")
    )
    os.makedirs(output_dir, exist_ok=True)

    for i in range(len(data_list)):
        # Normalize volume data
        if volume and volume_list is not None:
            volume_data = volume_list[i].squeeze()

            # Print original volume stats
            print(f"Original volume stats:")
            print(f"Min: {np.min(volume_data)}")
            print(f"Max: {np.max(volume_data)}")
            print(f"Mean: {np.mean(volume_data)}")

            # Reshape for StandardScaler
            volume_data_reshaped = volume_data.reshape(-1, 1)
            scaler = StandardScaler()
            volume_data_normalized = scaler.fit_transform(
                volume_data_reshaped
            ).squeeze()

            # Print normalized volume stats
            print(f"\nNormalized volume stats:")
            print(f"Min: {np.min(volume_data_normalized)}")
            print(f"Max: {np.max(volume_data_normalized)}")
            print(f"Mean: {np.mean(volume_data_normalized)}")

        # Calculate global min and max for amplitude axis
        global_min = np.min(data_list[i])
        global_max = np.max(data_list[i])

        # Add some padding to the limits
        y_range = global_max - global_min
        global_min -= y_range * 0.05
        global_max += y_range * 0.05

        # Individual channel analysis
        data = data_list[i].squeeze()

        # Get title and y-axis label from info_list
        title = info_list[i]["title"]
        y_axis = info_list[i]["y-axis"]

        # Create figure with subplots
        if volume and volume_list is not None:
            fig, (ax1, ax2) = plt.subplots(
                2,
                1,
                figsize=(12, 10),
                gridspec_kw={"height_ratios": [3, 1]},
                facecolor=colors["face"],
            )
            fig.subplots_adjust(hspace=0)
        else:
            fig = plt.figure(figsize=(12, 8), facecolor=colors["face"])
            ax1 = fig.add_subplot(111)

        # Plot price data
        x_indices = list(range(len(dates)))
        ax1.plot(x_indices, data, color=colors["line"], linewidth=1.2)

        # Title and labels for price subplot
        ax1.set_title(rf"\textbf{{{title}}}", fontsize=32, pad=20, color=colors["text"])
        ax1.set_ylabel(r"\textrm{" + y_axis + r"}", fontsize=26, color="black")
        ax1.set_ylim(global_min, global_max)

        # Set tick size for y-axis
        ax1.tick_params(axis="y", labelsize=26)

        if volume and volume_list is not None:
            # Remove x-axis labels from price subplot
            ax1.set_xticklabels([])
            ax1.set_xticks([])

            # Plot normalized volume bars
            ax2.bar(
                x_indices,
                volume_data_normalized,
                color="royalblue",
                alpha=1.0,
                width=1.0,
            )

            # Set y-axis limits for normalized volume
            ax2.set_ylim(
                min(volume_data_normalized) * 1.1, max(volume_data_normalized) * 1.1
            )

            # Style volume subplot
            ax2.set_ylabel(
                r"\textrm{Normalized Volume ($\sigma$)}", fontsize=26, color="black"
            )
            ax2.tick_params(axis="y", labelsize=26)

            # Remove x-axis
            ax2.set_xticklabels([])
            ax2.set_xticks([])

            # Style volume subplot spines
            for spine in ax2.spines.values():
                spine.set_color(colors["spine"])
                spine.set_linewidth(1.5)

            # Make grid more visible
            ax2.grid(True, axis="y", linestyle="-", alpha=0.3, color="gray")

            plt.subplots_adjust(bottom=0.1)
        else:
            ax1.set_xticklabels([])
            ax1.set_xticks([])

        # Style the price subplot spines
        for spine in ax1.spines.values():
            spine.set_color(colors["spine"])
            spine.set_linewidth(1.5)

        # Add subtle grid to price subplot
        ax1.grid(True, axis="y", linestyle="--", alpha=0.2, color=colors["grid"])

        # Save with tight layout
        figure_name = info_list[i]["file_codename"]
        plt.savefig(
            os.path.join(output_dir, f"{figure_name}.{output_format}"),
            dpi=dpi,
            bbox_inches="tight",
            facecolor=colors["face"],
            edgecolor="none",
        )
        plt.close()


def format_dates(dates):
    """
    Convert pandas datetime objects to 'MMM DD, YYYY (H:MMam)' format
    Example: Dec 06, 2023 (9:30am)

    Args:
        dates (pd.Series): Series of pandas datetime objects
    Returns:
        np.ndarray: Formatted date strings
    """
    # Create initial format with leading zeros
    formatted_dates = [
        d.strftime("%b %d, %Y (%I:%M%p)").replace(" 0", " ").lower() for d in dates
    ]

    # Capitalize month and handle day leading zeros
    capitalized = []
    for d in formatted_dates:
        # Split into parts
        month_part = d[:3].capitalize()  # First 3 chars are month
        rest = d[3:]  # Rest of the string

        # If day starts with 0, remove it (but keep it for single digit days)
        if rest.startswith(" 0"):
            rest = " " + rest[2:]

        capitalized.append(month_part + rest)

    return np.array(capitalized)


def plot_returns_distribution(data: np.ndarray, log: bool = False):
    """
    Plot the distribution of (optionally log) returns for a given security.
    """


# Example usage
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Time Series Analysis")
    parser.add_argument(
        "root", type=str, default="AAPL", help="Root of security to analyze."
    )
    parser.add_argument(
        "mode",
        type=str,
        default="option",
        help="Whether to analyze option or underlying data. Options: 'option', 'stock', or 'both'.",
    )
    parser.add_argument(
        "start_date",
        type=str,
        default="20241107",
        help="Start date for data in YYYYMMDD format.",
    )
    parser.add_argument(
        "end_date",
        type=str,
        default="20241107",
        help="End date for data in YYYYMMDD format.",
    )
    parser.add_argument(
        "exp",
        type=str,
        default="20250117",
        help="Expiration date for option in YYYYMMDD format.",
    )
    parser.add_argument(
        "strike", type=int, default=225, help="Strike price for option."
    )
    parser.add_argument(
        "right",
        type=str,
        default="C",
        help="Option type, either 'C' for call or 'P' for put.",
    )
    parser.add_argument(
        "output_format",
        type=str,
        default="eps",
        help="Output format for saving plots. Default is 'eps'.",
    )
    parser.add_argument(
        "--volume", action="store_true", help="Include volume data in analysis."
    )
    args = parser.parse_args()

    # Set plot parameters
    right_name = "Call" if args.right == "C" else "Put"
    map = {
        "option": {
            "title": args.root,
            "codename": args.root,
            "file_codename": f"{args.root}_{args.right}_{args.exp}_{args.strike}",
            "y-axis": f"{right_name} Option Midprice",
            "x-axis": "Date",
        },
        "stock": {
            "title": args.root,
            "codename": args.root,
            "file_codename": f"{args.root}_stock",
            "y-axis": "Underlying Midprice",
            "x-axis": "Date",
        },
    }

    from optrade.data.thetadata.contracts import Contract

    contract = Contract(
        root=args.root,
        start_date=args.start_date,
        exp=args.exp,
        strike=args.strike,
        interval_min=1,
        right=args.right,
    )

    df = load_all_data(contract=contract, clean_up=True)

    # Clip data to end datetime
    end_datetime = datetime.strptime(args.end_date, "%Y%m%d")
    df = df[df["datetime"] <= end_datetime]

    print(df.tail())

    data_list = []
    volume_list = []
    info_list = []

    if args.mode in ["option", "both"]:
        info_list.append(map["option"])
        option_mid = df["option_mid_price"].values
        data_list.append(option_mid)
        if args.volume:
            volume_list.append(df["option_volume"].values)

    if args.mode in ["stock", "both"]:
        info_list.append(map["stock"])
        stock_mid = df["stock_mid_price"].values
        data_list.append(stock_mid)
        if args.volume:
            volume_list.append(df["stock_volume"].values)

    dates = df["datetime"].values

    # In your main code, replace the ... with:
    dates = format_dates(df["datetime"])

    volume_list = []
    if args.volume:
        if args.mode in ["option", "both"]:
            volume_list.append(df["option_volume"].values)
        if args.mode in ["stock", "both"]:
            volume_list.append(df["stock_volume"].values)

    analyze_time_series(
        data_list=data_list,
        info_list=info_list,
        dates=dates,
        volume_list=None,
        dpi=600,
        output_format=args.output_format,
    )



import torch
import torch.nn as nn
from typing import List, Optional
from pydantic import BaseModel

# Layers and models
from optrade.torch.models.dlinear.model import DLinear
from optrade.torch.models.tsmixer.model import TSMixer

# Weight initialization
from optrade.torch.models.utils.weight_init import xavier_init

# Util Layers
from optrade.torch.models.utils.revin import RevIN
from optrade.torch.models.utils.patcher import Patcher


class EMForecaster(nn.Module):
    def __init__(
        self,
        args: BaseModel,
        seed: int = 42,
        seq_len: int = 336,
        pred_len: int = 96,
        num_channels: int = 1,
        revin: bool = True,
        revout: bool = True,
        revin_affine: bool = True,
        eps_revin: float = 1e-5,
        patch_model_id: str = "TSMixer",
        patch_norm: str = "none",
        patch_act: str = "GeLU",
        patch_dim: int = 24,
        patch_stride: int = 12,
        patch_embed_dim: int = 128,
        pos_enc: str = "none",
        return_head: bool = True,
        target_channels: Optional[list] = None,
    ) -> None:
        super(EMForecaster, self).__init__()

        """
        Patched-based forecasting model (univariate only).

        Args:
            seq_len (int): Length of input sequence.
            pred_len (int): Length of output forecasting.
            num_channels (int): Number of input channels (features).
            revin (bool): Whether to use reversible instance normalization (input).
            revout (bool): Whether to use reversible instance normalization (output).
            revin_affine (bool): Whether to use affine transformation for reversible instance normalization.
            eps_revin (float): Epsilon value for standard deviation numerical stability in reversible instance normalization.
            patch_model_id (str): Model used for patch embedding.
            patch_norm (str): Normalization layer used in patch embedding.
            patch_act (str): Activation function used in patch embedding.
            patch_dim (int): Patch dimension.
            patch_stride (int): Stride used in patching.
            patch_embed_dim (int): Patch embedding dimension.
            d_model (int): Dimension of the model.
            patching_on (bool): Only used for extensions of PatchedForecaster when patching is done externally.
            num_patches (int): Only used for extensions of PatchedForecaster when patching is done externally.
        """

        # Parameters
        self.args = args
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.num_channels = num_channels
        self.patch_embed_dim = patch_embed_dim
        self.patch_model_id = patch_model_id
        self.patch_dim = patch_dim
        self.return_head = return_head
        self.target_channels = target_channels
        self.revout = revout
        self.revin_affine = revin_affine
        self.eps_revin = eps_revin
        self.num_channels = num_channels

        if patch_stride == -1:
            self.patch_stride = self.patch_dim // 2
        elif patch_stride == -2:
            self.patch_stride = self.patch_dim
        else:
            self.patch_stride = patch_stride

        self.num_patches = int((seq_len - patch_dim) / self.patch_stride) + 2

        # RevIN
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        # Layers
        self.patcher = Patcher(self.patch_dim, self.patch_stride)
        self.patch_model = self.get_patch_model()

        # Positional encoding
        if pos_enc == "learnable":
            self.pos_enc = nn.Parameter(
                torch.randn(1, self.num_channels, self.num_patches * patch_embed_dim)
            )
        elif pos_enc == "none":
            self.pos_enc = 0

        # Activation function (patching)
        if patch_act == "relu":
            self.patch_act = nn.ReLU()
        elif patch_act == "gelu":
            self.patch_act = nn.GELU()
        else:
            self.patch_act = nn.Identity()

        # Fixed code:
        if target_channels is None:
            num_output_channels = self.num_channels
        else:
            num_output_channels = len(target_channels)

        self.head = nn.ModuleList(
            [
                nn.Linear(self.num_patches * self.patch_embed_dim, self.pred_len)
                for i in range(num_output_channels)
            ]
        )

        # Weight initialization
        self.apply(lambda m: xavier_init(m, seed=seed))

    def forward_patch_model(self, x):
        # Process each channel through its dedicated model
        outputs = []
        for i in range(len(self.patch_model)):
            channel_input = x[:, i, :, :]  # (batch_size, num_patches, patch_dim)
            channel_output = self.patch_model[i](
                channel_input
            )  # (batch_size, num_patches, patch_embed_dim)
            outputs.append(
                channel_output.view(-1, 1, self.num_patches, self.patch_embed_dim)
            )

        # Stack the outputs along the channel dimension
        x = torch.stack(
            outputs, dim=1
        )  # (batch_size, num_channels, num_patches, patch_embed_dim)

        return x

    def forward(self, x):

        # RevIN
        if self._revin:
            x = self.revin(x, mode="norm")

        # Patching
        x = self.patcher(
            x
        ).squeeze()  # (batch_size, num_channels, num_patches, patch_dim)

        # Activation function (optional)
        x = self.patch_act(x)  # (batch_size, num_channels, num_patches, patch_dim)

        # Patch model
        x = self.forward_patch_model(
            x
        )  # (batch_size, num_channels, num_patches, patch_embed_dim)

        if self.return_head:
            # Flatten
            x = x.view(
                -1, self.num_channels, self.num_patches * self.patch_embed_dim
            )  # (B, 1, num_patches * patch_embed_dim)

            # Positional Encoding
            x = (
                x + self.pos_enc
            )  # (B, self.num_channels, num_patches * patch_embed_dim)

            # Base Model + Linear Head
            outputs = []
            for i in range(len(self.head)):
                channel_input = x[
                    :, i, :
                ]  # (batch_size, num_patches * patch_embed_dim)
                channel_output = self.head[i](channel_input)  # (batch_size, pred_len)
                outputs.append(
                    channel_output.view(-1, 1, self.pred_len)
                )  # (batch_size, 1, pred_len)
            x = torch.stack(outputs, dim=1).squeeze(
                2
            )  # (batch_size, num_output_channels, pred_len)

            # RevOUT
            if self.revout:
                x = self.revin(x, mode="denorm")

        return x

        return map[self.patch_model_id]

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def get_patch_model(self) -> nn.Module:
        if self.patch_model_id == "DLinear":
            return nn.ModuleList(
                [
                    nn.Sequential(
                        DLinear(
                            task="forecasting",
                            seq_len=self.patch_dim,
                            pred_len=self.patch_embed_dim,
                            num_channels=self.num_patches,
                            moving_avg=self.args.emf.moving_avg,
                            individual=False,
                            return_head=False,
                        )
                    )
                    for i in range(self.num_channels)
                ]
            )
        elif self.patch_model_id == "TSMixer":
            return nn.ModuleList(
                [
                    nn.Sequential(
                        nn.Linear(self.patch_dim, self.patch_embed_dim),
                        TSMixer(
                            seq_len=self.patch_embed_dim,
                            pred_len=1,
                            num_enc_layers=self.args.emf.num_enc_layers,
                            d_model=self.args.emf.d_model,
                            num_channels=self.num_patches,
                            dropout=self.args.emf.dropout,
                            revin=False,
                            revin_affine=False,
                            revout=False,
                            return_head=False,
                        ),
                    )
                    for i in range(self.num_channels)
                ]
            )
        elif self.patch_model_id == "Linear":
            return NotImplementedError
        else:
            raise NotImplementedError


if __name__ == "__main__":
    from optrade.config.config import Global

    args = Global()

    # Multivariate Test
    batch_size = 32
    seq_len = 512
    pred_len = 96
    num_channels = 7
    x = torch.randn(batch_size, num_channels, seq_len)

    model = EMForecaster(
        args=args,
        seq_len=seq_len,
        pred_len=pred_len,
        num_channels=num_channels,
        patch_model_id="TSMixer",
        revin=True,
        revout=True,
        revin_affine=True,
        target_channels=[6, 3, 1],
    )

    y = model(x)
    print(f"Model output: {y.shape}")



from pathlib import Path
from typing import Tuple, List, Union, Optional
from rich.console import Console
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
import torch
from torch.utils.data import DataLoader, Dataset, ConcatDataset

# Data
from optrade.data.contracts import get_contract_datasets

# Datasets
from optrade.data.contracts import ContractDataset
from optrade.utils.error_handlers import (
    DataValidationError,
    INCOMPATIBLE_START_DATE,
    INCOMPATIBLE_END_DATE,
)

# Features
from optrade.data.features import transform_features

# Get absolute path for this script
SCRIPT_DIR = Path(__file__).resolve().parent


class ForecastingDataset(Dataset):
    """
    Args:
        data (torch.Tensor): Tensor of shape (num_time_steps, num_features)
        seq_len (int): Length of the lookback window
        pred_len (int): Length of the forecast window
        target_channels (list): Channels to forecast. By default target_channels=["option_returns"], which corresponds to the
                                option midprice returns. If None, all channels will be returned for the target.
        dtype (str): Desired data type of the tensor. Default is "float32".
    """

    def __init__(
        self,
        data: pd.DataFrame,
        seq_len: int,
        pred_len: int,
        target_channels: Optional[List[str]] = None,
        dtype: str = "float32",
    ) -> None:
        self.data = torch.tensor(data.to_numpy(), dtype=eval("torch." + dtype))
        self.seq_len = seq_len
        self.pred_len = pred_len

        if target_channels is not None and len(target_channels) > 0:
            feature_names = data.columns.to_list()
            self.target_channels_idx = [
                feature_names.index(channel) for channel in target_channels
            ]

    def __len__(self) -> int:
        return self.data.shape[0] - self.seq_len - self.pred_len

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Args:
            idx (int): Index of the starting point of the lookback window

        Returns:
            input (torch.Tensor): Lookback window of shape (num_features, seq_len)
            target (torch.Tensor): Target window of shape (len(target_channels), pred_len) if target_channels is not None, otherwise
                                   (num_features, pred_len).
        """
        input = self.data[idx : idx + self.seq_len]
        if self.target_channels:
            target = self.data[
                idx + self.seq_len : idx + self.seq_len + self.pred_len,
                self.target_channels_idx,
            ]
        else:
            target = self.data[idx + self.seq_len : idx + self.seq_len + self.pred_len]
        return input.transpose(0, 1), target.transpose(0, 1)


def normalize_concat_dataset(
    concat_dataset: ConcatDataset,
    scaler: StandardScaler,
) -> None:
    """
    Modifies the data in a ConcatDataset in-place by normalizing it using a fitted StandardScaler.

    Args:
        concat_dataset: ConcatDataset object containing ForecastingDatasets
        scaler: Fitted StandardScaler (scikit-learn)

    Returns:
        None
    """
    for dataset in concat_dataset.datasets:
        data = dataset.data.numpy()
        dtype = dataset.dtype

        # Normalize data
        normalized_data = scaler.transform(data)

        # Replace data with normalized version
        dataset.data = torch.tensor(normalized_data, dtype=dtype)


def normalize_datasets(
    train_dataset: ConcatDataset,
    val_dataset: ConcatDataset,
    test_dataset: ConcatDataset,
) -> Tuple[ConcatDataset, ConcatDataset, ConcatDataset, StandardScaler]:
    """
    Normalizes financial time series datasets using StandardScaler.
    Fits scaler only on training data to prevent look-ahead bias.

    Args:
        train_dataset: Training dataset (ConcatDataset of ForecastingDatasets)
        val_dataset: Validation dataset
        test_dataset: Test dataset
        save_scaler: Whether to save the scaler to disk
        scaler_path: Path to save the scaler if save_scaler is True

    Returns:
        Normalized train_dataset, val_dataset, test_dataset, and the fitted scaler
    """
    # Extract all underlying data from train_dataset
    all_train_data = []

    # Iterate through the individual ForecastingDataset objects within the ConcatDataset
    for dataset_idx in range(len(train_dataset.datasets)):
        individual_dataset = train_dataset.datasets[dataset_idx]

        # Extract data tensor from the individual dataset
        data = individual_dataset.data.numpy()
        all_train_data.append(data)

    # Stack all the arrays together
    combined_train_data = np.vstack(all_train_data)

    # Fit scaler on training data only
    scaler = StandardScaler()
    scaler.fit(combined_train_data)

    # Apply normalization to all datasets
    normalize_concat_dataset(train_dataset, scaler)
    normalize_concat_dataset(val_dataset, scaler)
    normalize_concat_dataset(test_dataset, scaler)

    return train_dataset, val_dataset, test_dataset, scaler


def get_forecasting_dataset(
    contracts: ContractDataset,
    seq_len: int,
    pred_len: int,
    tte_tolerance: Tuple[int, int],
    core_feats: List[str] = ["option_returns"],
    tte_feats: Optional[List[str]] = None,
    datetime_feats: Optional[List[str]] = None,
    clean_up: bool = True,
    offline: bool = False,
    intraday: bool = False,
    target_channels: Optional[List[str]] = None,
    dtype: str = "float32",
    save_dir: Optional[str] = None,
    download_only: bool = False,
    verbose: bool = False,
) -> Dataset:
    """
    Creates a PyTorch dataset object composed of multiple ForecastingDatasets, each representing
    different option contracts.

    Args:
        contracts: ContractDataset object containing option contract parameters
        core_feats: List of core features to include
        tte_feats: List of time-to-expiration features to include
        datetime_feats: List of datetime features to include
        tte_tolerance: Tuple of (min, max) time to expiration tolerance in days
        clean_up: Whether to clean up the data after use
        offline: Whether to load saved contracts from disk
        intraday: Whether to use intraday data
        target_channels: List of target channels to include in the target tensor. If None, all channels will be included.
        seq_len: Sequence length of lookback window (input)
        pred_len: Prediction length of forecast window (target)
        dtype: Data type for the PyTorch tensors
        save_dir: Save directory
        download_only: Whether to download data only (used mainly for Universe class)
        verbose: Whether to print verbose output

    Returns:
        Concatenated PyTorch dataset object
    """

    ctx = Console()
    dataset_list = []

    if download_only:
        clean_up = offline = False

    for contract in contracts.contracts:
        # Flag to track if we should try the next contract
        move_to_next_contract = False

        # Continue trying the current contract until we succeed or decide to move on
        while not move_to_next_contract:
            try:
                df = contract.load_data(
                    save_dir=save_dir, clean_up=clean_up, offline=offline
                )

                if not download_only:
                    # Select and add features
                    data = transform_features(
                        df=df,
                        core_feats=core_feats,
                        tte_feats=tte_feats,
                        datetime_feats=datetime_feats,
                        strike=contract.strike,
                        exp=contract.exp,
                    )

                    # Convert to PyTorch dataset
                    dataset = ForecastingDataset(
                        data=data,
                        seq_len=seq_len,
                        pred_len=pred_len,
                        target_channels=target_channels,
                        dtype=dtype,
                    )
                    dataset_list.append(dataset)
                else:
                    pass

                # If we get here successfully, move to the next contract
                move_to_next_contract = True

            except DataValidationError as e:
                if e.error_code == INCOMPATIBLE_START_DATE:
                    new_start_date = e.real_start_date

                    # Check if (exp - new_start_date) is within tte_tolerance. If not, move to the next contract
                    if pd.to_datetime(contract.exp, format="%Y%m%d") - pd.to_datetime(
                        new_start_date, format="%Y%m%d"
                    ) < pd.Timedelta(days=tte_tolerance[0]):
                        if verbose:
                            ctx.log(
                                f"Option contract start date mismatch. New start date {new_start_date} is too close to expiration {contract.exp}. Moving to next contract."
                            )
                        move_to_next_contract = True
                    # If tte_tolerance is satisfied, update the contract with the new start date and try again
                    else:
                        if verbose:
                            ctx.log(
                                f"Option contract start date mismatch. Attempting to get data for {contract} with new start date: {new_start_date}"
                            )
                        contract.start_date = new_start_date
                elif e.error_code == INCOMPATIBLE_END_DATE:
                    new_start_date = e.real_start_date
                    new_exp = e.real_end_date

                    # Check if (new_exp - new_start_date) is within tte_tolerance. If not, move to the next contract
                    if pd.to_datetime(new_exp, format="%Y%m%d") - pd.to_datetime(
                        new_start_date, format="%Y%m%d"
                    ) < pd.Timedelta(days=tte_tolerance[0]):
                        if verbose:
                            ctx.log(
                                f"Option contract start date and end date mismatch. New start date {new_start_date} is too close to new expiration {new_exp}. Moving to next contract."
                            )
                        move_to_next_contract = True
                    # If tte_tolerance is satisfied, update contract expiration to the observed end date of option data
                    else:
                        # For other DataValidationError types, move to the next contract
                        if verbose:
                            ctx.log(
                                f"DataValidationError for {contract}: {e}. Moving to next contract."
                            )
                        move_to_next_contract = True
                        contract.exp = new_exp
                else:
                    # For other DataValidationError types, move to the next contract
                    if verbose:
                        ctx.log(
                            f"DataValidationError for {contract}: {e}. Moving to next contract."
                        )
                    move_to_next_contract = True
            except Exception as e:
                if verbose:
                    ctx.log(
                        f"Unknown error for {contract}: {e}. Moving to next contract."
                    )
                move_to_next_contract = True

    if download_only:
        return
    else:
        return ConcatDataset(dataset_list)


def get_forecasting_loaders(
    root: str = "AAPL",
    start_date: str = "20231107",
    end_date: str = "20241114",
    contract_stride: int = 5,
    interval_min: int = 1,
    right: str = "C",
    target_tte: int = 30,
    tte_tolerance: Tuple[int, int] = (25, 35),
    moneyness: str = "OTM",
    target_band: float = 0.05,
    volatility_type: str = "period",
    volatility_scaled: bool = True,
    volatility_scalar: float = 1.0,
    train_split: float = 0.7,
    val_split: float = 0.1,
    core_feats: List[str] = ["option_returns"],
    tte_feats: List[str] = ["sqrt"],
    datetime_feats: List[str] = ["sin_timeofday"],
    batch_size: int = 32,
    shuffle: bool = True,
    drop_last: bool = False,
    num_workers: int = 4,
    prefetch_factor: int = 2,
    pin_memory: bool = torch.cuda.is_available(),
    clean_up: bool = True,
    offline: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
    scaling: bool = True,
    intraday: bool = False,
    target_channels: List[int] = [0],
    seq_len: int = 100,
    pred_len: int = 10,
    dtype: str = "float64",
) -> Union[
    Tuple[DataLoader, DataLoader, DataLoader],
    Tuple[DataLoader, DataLoader, DataLoader, StandardScaler],
]:
    """

    Forms training, validation, and test dataloaders for option contract data.

    Args:
        root: Underlying stock symbol
        start_date: Start date for the total dataset in YYYYMMDD format
        end_date: End date for the total dataset in YYYYMMDD format
        contract_stride: Number of days between each contract (for sampling)
        interval_min: Interval in minutes for the underlying stock data and option data
        right: Option type (C for call, P for put)
        target_tte: Target time to expiration in days
        tte_tolerance: Tuple of (min, max) time to expiration tolerance in minutes
        moneyness: Moneyness of the option contract (OTM, ATM, ITM)
        target_band: Target band for moneyness selection
        volatility_type: Type of historical volatility to use
        volatility_scaled: Whether to scale strikes based on historical volatility
        volatility_scalar: Scalar to adjust historical volatility
        train_split: Proportion of total days to use for training
        val_split: Proportion of total days to use for validation
        core_feats: List of core features to include
        tte_feats: List of time-to-expiration features to include
        datetime_feats: List of datetime features to include
        batch_size: Number of samples per batch
        shuffle: Whether to shuffle the data
        drop_last: Whether to drop the last incomplete batch
        num_workers: Number of subprocesses to use for data loading
        prefetch_factor: Number of batches to prefetch
        pin_memory: Whether to pin memory for faster GPU transfer
        clean_up: Whether to clean up the data after use
        offline: Whether to load saved contracts from disk
        contract_dir: Directory to save/load contracts
        verbose: Whether to print verbose output
        scaling: Whether to normalize the datasets

    Returns:
        Training, validation, and test PyTorch dataloaders.
    """

    train_contracts, val_contracts, test_contracts = get_contract_datasets(
        root=root,
        start_date=start_date,
        end_date=end_date,
        contract_stride=contract_stride,
        interval_min=interval_min,
        right=right,
        target_tte=target_tte,
        tte_tolerance=tte_tolerance,
        moneyness=moneyness,
        target_band=target_band,
        volatility_type=volatility_type,
        volatility_scaled=volatility_scaled,
        volatility_scalar=volatility_scalar,
        train_split=train_split,
        val_split=val_split,
        clean_up=clean_up,
        offline=offline,
        save_dir=save_dir,
        verbose=verbose,
    )

    # Get the combined datasets of contract data for training, validation, and testing
    train_dataset = get_forecasting_dataset(
        contracts=train_contracts,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        tte_tolerance=tte_tolerance,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        seq_len=seq_len,
        pred_len=pred_len,
        dtype=dtype,
        verbose=verbose,
    )

    val_dataset = get_forecasting_dataset(
        contracts=val_contracts,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        tte_tolerance=tte_tolerance,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        seq_len=seq_len,
        pred_len=pred_len,
        dtype=dtype,
        verbose=verbose,
    )
    test_dataset = get_forecasting_dataset(
        contracts=test_contracts,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        tte_tolerance=tte_tolerance,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        seq_len=seq_len,
        pred_len=pred_len,
        dtype=dtype,
        verbose=verbose,
    )

    if scaling:
        train_dataset, val_dataset, test_dataset, scaler = normalize_datasets(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            test_dataset=test_dataset,
        )
    else:
        scaler = None

    # Create dataloaders for training, validation, and testing
    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=True,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        dataset=val_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=True,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )
    test_loader = DataLoader(
        dataset=test_dataset,
        batch_size=batch_size,
        shuffle=False,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=True,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )

    return train_loader, val_loader, test_loader, scaler


def create_windows(
    df: pd.DataFrame,
    seq_len: int,
    pred_len: int,
    window_stride: int,
    intraday: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates rolling windows of data for a given DataFrame. Should be used primarily for scikit-learn models and/or intraday modeling,
    otherwise default to optrade.data.forecasing.get_forecasting_loaders or optrade.data.forecasting.get_forecasting_datasets.

    Args:
        df (pd.DataFrame): DataFrame containing the data.
        seq_len (int): Length of the input sequence.
        pred_len (int): Length of the prediction sequence.
        window_stride (int): Number of steps to move the window forward.
        intraday (bool): Whether the data is intraday or not. If True, the function will first split the data
                         into separate trading days before creating individual windows that cannot crossover
                         between days. Otherwise, the function will create windows that can span multiple days.
    Returns:
        input (np.ndarray): Array of input windows of shape (num_windows, seq_len, num_features) where num_features
                            is the number of columns in the DataFrame (removing datetime but adding returns).
        target (np.ndarray): Array of target windows of shape (num_windows, pred_len, 1).
                             Target contains only returns for the 'option_mid_price'.
    """
    datetime = df["datetime"]
    df_copy = df.copy()

    # Define input features (all columns except datetime)
    feature_columns = [col for col in df_copy.columns if col != "datetime"]
    inputs, targets = [], []

    if intraday:
        # Get all unique days
        days = datetime.dt.date.unique()

        # Iterate through each day independently
        for day in days:
            day_mask = datetime.dt.date == day
            day_data = df_copy.loc[day_mask].copy()

            print(f"Length of data: {len(day_data)}")

            # Since returns will be part of the input features but don't exist for 9:30am
            # we remove the market open (9:30am) of each day
            first_time = day_data["datetime"].iloc[0].time()
            if first_time.hour == 9 and first_time.minute == 30:
                day_data = day_data.iloc[1:].reset_index(drop=True)
                print(f"Day data after removing 9:30am: {day_data.tail()}")

            # Raise an error if the length of the day is less than the sum of seq_len+pred_len
            if len(day_data) < seq_len + pred_len:
                raise ValueError(
                    f"seq_len + pred_len = {seq_len + pred_len} exceeds the length of the day. \
                    Either set intraday=False or reduce seq_len and/or pred_len."
                )

            # Get input features and targets, convert to NumPy arrays
            day_features = day_data[feature_columns].to_numpy()
            day_targets = day_data["option_returns"].to_numpy().reshape(-1, 1)

            # Apply the sliding window technique to obtain windows for inputs and targets
            for i in range(0, len(day_data) - seq_len - pred_len + 1, window_stride):
                inputs.append(day_features[i : i + seq_len])
                targets.append(day_targets[i + seq_len : i + seq_len + pred_len])
    else:
        # Since returns will be part of the input features but don't exist for first market open
        # i.e. 9:30am on the first day, we remove it
        first_time = datetime.iloc[0].time()
        if first_time.hour == 9 and first_time.minute == 30:
            df_copy = df_copy.iloc[1:].reset_index(drop=True)

        # Extract features and targets
        features = df_copy[feature_columns].to_numpy()
        targets_data = df_copy["option_returns"].to_numpy().reshape(-1, 1)

        # Create windows
        for i in range(0, len(df_copy) - seq_len - pred_len + 1, window_stride):
            inputs.append(features[i : i + seq_len])
            targets.append(targets_data[i + seq_len : i + seq_len + pred_len])

    # Convert to numpy arrays
    return np.array(inputs), np.array(targets)


if __name__ == "__main__":
    # Test: get_forecasting_loaders
    root = "AMZN"
    total_start_date = "20230101"
    total_end_date = "20230601"
    right = "C"
    interval_min = 60
    contract_stride = 5
    target_tte = 30
    tte_tolerance = (15, 45)
    moneyness = "ATM"
    volatility_scaled = True
    volatility_scalar = 0.01
    volatility_type = "period"
    target_band = 0.05

    # TTE features
    tte_feats = ["sqrt", "exp_decay"]

    # Datetime features
    datetime_feats = [
        "sin_minute_of_day",
        "cos_minute_of_day",
        "sin_hour_of_week",
        "cos_hour_of_week",
    ]

    # Select features
    core_feats = [
        "option_returns",
        "stock_returns",
        "distance_to_strike",
        "moneyness",
        "option_lob_imbalance",
        "option_quote_spread",
        "stock_lob_imbalance",
        "stock_quote_spread",
        "option_mid_price",
        "option_bid_size",
        "option_bid",
        "option_ask_size",
        "option_close",
        "option_volume",
        "option_count",
        "stock_mid_price",
        "stock_bid_size",
        "stock_bid",
        "stock_ask_size",
        "stock_ask",
        "stock_volume",
        "stock_count",
    ]

    # Testing: get_loaders
    output = get_forecasting_loaders(
        root=root,
        start_date=total_start_date,
        end_date=total_end_date,
        contract_stride=contract_stride,
        interval_min=interval_min,
        right=right,
        target_tte=target_tte,
        tte_tolerance=tte_tolerance,
        moneyness=moneyness,
        target_band=target_band,
        volatility_type=volatility_type,
        volatility_scaled=volatility_scaled,
        volatility_scalar=volatility_scalar,
        train_split=0.4,
        val_split=0.3,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        batch_size=32,
        clean_up=False,
        offline=False,
        save_dir=None,
        verbose=True,
        scaling=True,
    )
    train_loader, val_loader, test_loader = output[0:3]

    print(f"Num train examples: {len(train_loader.dataset)}")
    print(f"Num val examples: {len(val_loader.dataset)}")
    print(f"Num test examples: {len(test_loader.dataset)}")

    # Testing: create_windows
    from optrade.data.features import transform_features
    from optrade.data.contracts import Contract
    from rich.console import Console

    console = Console()

    contract = Contract.find_optimal(
        root="AAPL",
        start_date="20241107",
        volatility_scaled=False,
        target_band=0.05,
        moneyness="OTM",
        interval_min=1,
        right="C",
    )

    df = contract.load_data(clean_up=True, offline=False)

    # TTE features
    tte_feats = ["sqrt", "exp_decay"]

    # Datetime features
    datetime_feats = [
        "sin_minute_of_day",
        "cos_minute_of_day",
        "sin_hour_of_week",
        "cos_hour_of_week",
    ]

    # Select features
    core_feats = [
        "option_returns",
        "stock_returns",
        "distance_to_strike",
        "moneyness",
        "option_lob_imbalance",
        "option_quote_spread",
        "stock_lob_imbalance",
        "stock_quote_spread",
        "option_mid_price",
        "option_bid_size",
        "option_bid",
        "option_ask_size",
        "option_close",
        "option_volume",
        "option_count",
        "stock_mid_price",
        "stock_bid_size",
        "stock_bid",
        "stock_ask_size",
        "stock_ask",
        "stock_volume",
        "stock_count",
    ]

    df = transform_features(
        df=df,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        strike=contract.strike,
        exp=contract.exp,
        keep_datetime=True,
    )
    print(df.columns)

    x, y = create_windows(
        df=df, seq_len=30, pred_len=6, window_stride=1, intraday=False
    )

    print(x.shape, y.shape)



import torch
import torch.nn as nn

from typing import Optional


class RevIN(nn.Module):
    def __init__(
        self,
        num_channels: int,
        eps: float = 1e-5,
        affine: bool = True,
        target_channels: Optional[list] = None,
    ) -> None:
        """
        Kim et al. (2022): Reversible instance normalization for accurate time-series forecasting against
        distribution shift. Provides a learnable instance normalization layer that is reversible. Code is
        from https://github.com/yuqinie98/PatchTST/blob/main/PatchTST_supervised/layers/RevIN.py with modifications,
        which was originally taken from https://github.com/ts-kim/RevIN.

        Args:
            num_channels: The number of features or channels.
            eps: A value added for numerical stability.
            affine: If True, RevIN has learnable affine parameters (e.g., like LayerNorm).
            target_channels: List of target channels for the head layer. It not None, it will denormalize only the target channels corresponding
                             to an input shape of (batch_size, len(target_channels), pred_len) for the model output.
        """

        super(RevIN, self).__init__()
        self.num_channels = num_channels
        self.eps = eps
        self.affine = affine

        if target_channels is not None:
            self.target_channels = target_channels
            self.affine_weight = nn.Parameter(torch.ones(len(target_channels)))
            self.affine_bias = nn.Parameter(torch.zeros(len(target_channels)))
        else:
            self.target_channels = None
            self.affine_weight = nn.Parameter(torch.ones(self.num_channels))
            self.affine_bias = nn.Parameter(torch.zeros(self.num_channels))

    def forward(self, x, mode: str):
        """
        Forward pass for normalization or denormalizating the time series with learnable affine transformations.

        Args:
            x: Input tensor of shape (batch_size, num_channels, seq_len).
            mode: 'norm' for normalization and 'denorm' for denormalization.
        Returns:
            Normalized or denormalized tensor of same shape as input tensor.
        """

        if x.dim() != 3:
            x = x.unsqueeze(1)

        if mode == "norm":
            self._get_statistics(x)
            x = self._normalize(x)
        elif mode == "denorm":
            x = self._denormalize(x)
        else:
            raise NotImplementedError
        return x

    def _get_statistics(self, x):
        self.mean = torch.mean(x, dim=2, keepdim=True).detach()
        self.stdev = torch.sqrt(
            torch.var(x, dim=2, keepdim=True, unbiased=False) + self.eps
        ).detach()

    def _normalize(self, x):
        x = x - self.mean
        x = x / self.stdev
        if self.affine:
            if self.target_channels is not None:
                x[:, self.target_channels, :] = x[
                    :, self.target_channels, :
                ] * self.affine_weight.unsqueeze(0).unsqueeze(
                    -1
                )  # Reshape: (1, num_channels, 1)
                x[:, self.target_channels, :] = x[
                    :, self.target_channels, :
                ] + self.affine_bias.unsqueeze(0).unsqueeze(
                    -1
                )  # Reshape: (1, num_channels, 1)
            else:
                x = x * self.affine_weight.unsqueeze(0).unsqueeze(
                    -1
                )  # Reshape: (1, num_channels, 1)
                x = x + self.affine_bias.unsqueeze(0).unsqueeze(
                    -1
                )  # Reshape: (1, num_channels, 1)
        return x

    def _denormalize(self, x):
        if self.affine:
            x = x - self.affine_bias.unsqueeze(0).unsqueeze(-1)
            x = x / (
                self.affine_weight.unsqueeze(0).unsqueeze(-1) + self.eps * self.eps
            )

        if self.target_channels is not None:
            x = x * self.stdev[:, self.target_channels, :]
            x = x + self.mean[:, self.target_channels, :]
        else:
            x = x * self.stdev
            x = x + self.mean

        return x


if __name__ == "__main__":
    batch_size = 32
    num_channels = 7
    seq_len = 100
    pred_len = 48
    target_channels = [1, 3]
    x = torch.randn(batch_size, num_channels, seq_len)
    revin = RevIN(
        num_channels=num_channels,
        eps=1e-5,
        affine=True,
        target_channels=target_channels,
    )
    x_norm = revin(x, mode="norm")

    print(f"x_norm shape: {x_norm.shape}")

    y_hat = torch.randn(batch_size, len(target_channels), pred_len)

    x_denorm = revin(y_hat, mode="denorm")

    print(f"x_denorm shape: {x_denorm.shape}")



from enum import Enum


# Define enums for categorical factors
class ThreeFactorLevel(str, Enum):
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class FiveFactorLevel(str, Enum):
    VERY_LOW = "very_low"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    VERY_HIGH = "very_high"


class SectorType(str, Enum):
    TECH = "technology"
    HEALTHCARE = "healthcare"
    FINANCIAL = "financial"
    CONSUMER_CYCLICAL = "consumer_cyclical"
    CONSUMER_DEFENSIVE = "consumer_defensive"
    INDUSTRIAL = "industrial"
    ENERGY = "energy"
    MATERIALS = "materials"
    UTILITIES = "utilities"
    REAL_ESTATE = "real_estate"
    COMMUNICATION = "communication"


class IndustryType(str, Enum):
    # Technology Industries
    SEMICONDUCTOR = "semiconductor"
    SOFTWARE_APPLICATION = "software_application"
    SOFTWARE_INFRASTRUCTURE = "software_infrastructure"
    ELECTRONIC_COMPONENTS = "electronic_components"
    COMPUTER_HARDWARE = "computer_hardware"
    CONSUMER_ELECTRONICS = "consumer_electronics"

    # Healthcare Industries
    BIOTECHNOLOGY = "biotechnology"
    PHARMACEUTICAL_MANUFACTURERS = "pharmaceutical_manufacturers"
    MEDICAL_DEVICES = "medical_devices"
    MEDICAL_CARE_FACILITIES = "medical_care_facilities"
    HEALTH_INFORMATION_SERVICES = "health_information_services"

    # Financial Industries
    BANKS_REGIONAL = "banks_regional"
    BANKS_DIVERSIFIED = "banks_diversified"
    INSURANCE_PROPERTY_CASUALTY = "insurance_property_casualty"
    INSURANCE_LIFE = "insurance_life"
    ASSET_MANAGEMENT = "asset_management"
    CAPITAL_MARKETS = "capital_markets"

    # Consumer Cyclical Industries
    AUTO_MANUFACTURERS = "auto_manufacturers"
    AUTO_PARTS = "auto_parts"
    APPAREL_RETAIL = "apparel_retail"
    DEPARTMENT_STORES = "department_stores"
    RESTAURANTS = "restaurants"
    LODGING = "lodging"
    TRAVEL_SERVICES = "travel_services"

    # Consumer Defensive Industries
    GROCERY_STORES = "grocery_stores"
    DISCOUNT_STORES = "discount_stores"
    HOUSEHOLD_PRODUCTS = "household_products"
    BEVERAGES_NON_ALCOHOLIC = "beverages_non_alcoholic"
    BEVERAGES_ALCOHOLIC = "beverages_alcoholic"
    TOBACCO = "tobacco"

    # Industrial Industries
    AEROSPACE_DEFENSE = "aerospace_defense"
    FARM_HEAVY_MACHINERY = "farm_heavy_machinery"
    ENGINEERING_CONSTRUCTION = "engineering_construction"
    TOOLS_ACCESSORIES = "tools_accessories"
    BUSINESS_EQUIPMENT = "business_equipment"

    # Energy Industries
    OIL_GAS_INTEGRATED = "oil_gas_integrated"
    OIL_GAS_EXPLORATION_PRODUCTION = "oil_gas_exploration_production"
    OIL_GAS_REFINING_MARKETING = "oil_gas_refining_marketing"
    OIL_GAS_EQUIPMENT_SERVICES = "oil_gas_equipment_services"

    # Materials Industries
    CHEMICALS = "chemicals"
    SPECIALTY_CHEMICALS = "specialty_chemicals"
    STEEL = "steel"
    BUILDING_MATERIALS = "building_materials"
    PAPER_PAPER_PRODUCTS = "paper_paper_products"

    # Utilities Industries
    UTILITIES_REGULATED_ELECTRIC = "utilities_regulated_electric"
    UTILITIES_REGULATED_GAS = "utilities_regulated_gas"
    UTILITIES_RENEWABLE = "utilities_renewable"
    UTILITIES_DIVERSIFIED = "utilities_diversified"

    # Real Estate Industries
    REIT_RESIDENTIAL = "reit_residential"
    REIT_RETAIL = "reit_retail"
    REIT_OFFICE = "reit_office"
    REIT_HEALTHCARE = "reit_healthcare"
    REIT_INDUSTRIAL = "reit_industrial"
    REAL_ESTATE_SERVICES = "real_estate_services"

    # Communication Industries
    TELECOM_SERVICES = "telecom_services"
    ENTERTAINMENT = "entertainment"
    ADVERTISING_AGENCIES = "advertising_agencies"
    ELECTRONIC_GAMING_MULTIMEDIA = "electronic_gaming_multimedia"
    INTERNET_CONTENT_INFORMATION = "internet_content_information"



import os
import torch


class EarlyStopping:
    """
    Early stops the training if validation loss doesn't improve after a given patience.

    Github:
        https://github.com/Bjarten/early-stopping-pytorch/blob/master/pytorchtools.py

    """

    def __init__(self, patience=7, verbose=True, delta=0, path="checkpoint.pt") -> None:
        """
        Args:
            patience (int): How long to wait after last time validation loss improved.
                            Default: 7
            verbose (bool): If True, prints a message for each validation loss improvement.
                            Default: False
            path (str): Path for the checkpoint to be saved to.
                            Default: 'checkpoint.pt'
        """
        self.patience = patience
        self.verbos = verbose
        self.counter = 0
        self.best_val_loss = float("inf")
        self.early_stop = False
        self.path = path

    def __call__(self, val_loss: float, model: torch.nn.Module) -> None:

        if val_loss > self.best_val_loss:
            self.counter += 1
            print(f"EarlyStopping counter: {self.counter} out of {self.patience}")
            if self.counter >= self.patience:
                self.early_stop = True
        else:
            self.save_checkpoint(val_loss, model)
            self.counter = 0

    def save_checkpoint(self, val_loss: float, model: torch.nn.Module) -> None:

        if self.verbos:
            print(
                f"Validation loss decreased ({self.best_val_loss:.6f} --> {val_loss:.6f})."
            )
        path_dir = os.path.abspath(os.path.dirname(self.path))

        if not os.path.isdir(path_dir):
            os.makedirs(path_dir)

        print(f"Saving Model Weights in {self.path}...")
        torch.save(model.state_dict(), self.path)
        self.best_val_loss = val_loss



import torch
import torch.nn as nn
import torch.nn.init as init


class PositionalEncoding(nn.Module):
    def __init__(self, patch_dim: int = 16, d_model: int = 128, num_patches: int = 64):
        super(PositionalEncoding, self).__init__()
        self.projection = nn.Linear(patch_dim, d_model)  # P x D projection matrix
        self.pos_encoding = nn.Parameter(
            torch.empty(num_patches, d_model)
        )  # N x D positional encoding matrix

        # Weight initialization
        init.xavier_uniform_(self.projection.weight)
        init.uniform_(self.pos_encoding, -0.02, 0.02)

    def forward(self, x):
        """
        Args:
            x: input tensor of shape (B, M, N, P) where B = batch_size, M = num_channels, N = num_patches,
               P = patch_dim.
        Returns:
            x: tensor of shape (B, M, N, D) where D = d_model.
        """

        B, M, N, P = x.shape
        x = x.view(
            B * M, N, P
        )  # Reshape the tensor to (B * M, N, P). We process each channel independently.
        x = self.projection(x) + self.pos_encoding.unsqueeze(0)
        x = x.view(B, M, N, -1)  # Reshape the tensor to (B, M, N, D).

        return x


# # Test PatchtstOG vs ours
# # Hyperparameters
# batch_size, num_channels, num_patches, patch_dim = 3, 4, 17, 6
# d_model = 5
# x = torch.randn(batch_size, num_channels, num_patches, patch_dim)

# # #==========================================MANUAL======================================================

# # Define parameters
# W_P_og = nn.Linear(patch_dim, d_model)
# W_pos_og = torch.empty((num_patches, d_model))

# # Initalization for W_P and W_pos
# nn.init.xavier_uniform_(W_P_og.weight)
# nn.init.uniform_(W_pos_og, -0.02, 0.02)


# # Fix random parameters
# # Print out parameters (out_features, in_features) format
# print(f"OG weight shape {W_P_og.weight.data.shape}")
# print(f"OG bias shape {W_P_og.bias.data.shape}")

# W_P_og.weight.data = torch.randn(d_model, patch_dim) * 5
# W_P_og.bias.data = torch.randn(d_model) * 17.3

# print(f"OG pos enc shape {W_pos_og.data.shape}")
# W_pos_og.data = torch.randn(num_patches, d_model) * (-2.4)


# #==========================================OURS======================================================


# # Initialize matrices to the same values
# positional_encoding = PositionalEncoding(patch_dim=patch_dim, d_model=d_model, num_patches=num_patches)

# print(f"(Before update) Ours weight shape {positional_encoding.projection.weight.data.shape}, "
#       f"bias shape {positional_encoding.projection.bias.data.shape}, "
#       f"and pos enc shape {positional_encoding.pos_encoding.data.shape}")

# positional_encoding.projection.weight.data = W_P_og.weight.data
# positional_encoding.projection.bias.data = W_P_og.bias.data
# positional_encoding.pos_encoding.data = W_pos_og.data

# print(f"(After update) Ours weight shape {positional_encoding.projection.weight.data.shape}, "
#       f"bias shape {positional_encoding.projection.bias.data.shape}, "
#       f"and pos enc shape {positional_encoding.pos_encoding.data.shape}")


# # OG
# # Project + Positional Encoding
# print(f"x shape: {x.shape}")
# print(f"W_P_og shape: {W_P_og.weight.data.shape}")
# u = W_P_og(x)                                                          # x: [bs x nvars x patch_num x d_model]
# u = torch.reshape(u, (u.shape[0]*u.shape[1],u.shape[2],u.shape[3]))      # u: [bs * nvars x patch_num x d_model]
# u = u + W_pos_og                                        # u: [bs * nvars x patch_num x d_model]

# # OURS
# print(u.shape)

# v = positional_encoding(x)
# print(v.shape)
# assert torch.allclose(u, v.view(batch_size*num_channels, num_patches, -1)), "Error: Positional Encoding not equal to OG"



import os
import shutil
from pathlib import Path
from typing import Tuple, Optional


def clean_up_file(file_path: str) -> None:
    try:
        os.remove(file_path)
    except OSError as e:
        print(f"Warning: Could not delete file {file_path}: {e}")


def clean_up_dir(dir_path: str) -> None:
    """
    Remove a directory and all its contents (files and subdirectories).

    Args:
        dir_path (str): Path to the directory to be removed
    """
    try:
        shutil.rmtree(dir_path)
    except OSError as e:
        print(f"Warning: Could not delete directory {dir_path}: {e}")


def set_contract_dir(
    SCRIPT_DIR: Path,
    root: str = "AAPL",
    start_date: str = "20231107",
    end_date: str = "20241114",
    contract_stride: int = 5,
    interval_min: int = 1,
    right: str = "C",
    target_tte: int = 30,
    tte_tolerance: Tuple[int, int] = (25, 35),
    moneyness: str = "OTM",
    target_band: float = 0.05,
    volatility_scaled: bool = True,
    volatility_scalar: float = 1.0,
    hist_vol: Optional[float] = None,
    save_dir: Optional[str] = None,
) -> Path:
    # Directory setup
    if save_dir is None:
        save_dir = SCRIPT_DIR.parent / "historical_data" / "contracts"
    else:
        save_dir = Path(save_dir) / "contracts"

    # Create a structured path based on key parameters
    contract_dir = (
        save_dir
        / root
        / f"{start_date}_{end_date}"
        / right
        / f"contract_stride_{contract_stride}"
        / f"interval_{interval_min}"
        / f"target_tte_{target_tte}"
        / f"moneyness_{moneyness}"
    )

    # Add volatility info to path if volatility_scaled is True
    if volatility_scaled:
        contract_dir = (
            contract_dir / f"histvol_{hist_vol}_volscalar_{volatility_scalar}"
        )
    else:
        contract_dir = (
            contract_dir / f"target_band_{str(target_band).replace('.', 'p')}"
        )

    return contract_dir



import torch
import torch.nn as nn
from typing import Optional

from optrade.torch.models.utils.revin import RevIN
from optrade.torch.models.utils.weight_init import xavier_init
from optrade.torch.models.utils.utils import Reshape


class Linear(nn.Module):
    def __init__(
        self,
        seq_len,
        pred_len,
        num_channels,
        norm_mode="layer",
        revin=True,
        revout=False,
        revin_affine=False,
        eps_revin=1e-5,
        channel_independent: bool = False,
        target_channels: Optional[list] = None,
    ) -> None:
        super(Linear, self).__init__()

        # Normalization
        self.d_model = num_channels
        self.num_channels = num_channels
        self.target_channels = target_channels
        self.revout = revout
        self.revin_affine = revin_affine
        self.eps_revin = eps_revin
        self.channel_independent = channel_independent

        # RevIN
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        if channel_independent:
            self.backbone = nn.Linear(seq_len, pred_len)
        else:
            self.backbone = nn.ModuleList(
                [
                    Reshape(-1, num_channels * seq_len),
                    nn.Linear(num_channels * seq_len, num_channels * pred_len),
                    Reshape(-1, num_channels, pred_len),
                ]
            )

        if target_channels is not None:
            self.target_channels = target_channels

            if channel_independent:
                self.head = nn.Linear(pred_len, pred_len)
            else:
                self.head = nn.ModuleList(
                    [
                        Reshape(-1, len(target_channels) * pred_len),
                        nn.Linear(
                            len(target_channels) * pred_len,
                            len(target_channels) * pred_len,
                        ),
                        Reshape(-1, len(target_channels), pred_len),
                    ]
                )

        # Weight initialization
        self.apply(xavier_init)

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def forward(self, x):

        # RevIN
        if self._revin:
            x = self.revin(x, mode="norm")

        if not self.channel_independent:
            for module in self.backbone:
                x = module(x)
            out = x
        else:
            out = self.backbone(x)  # (batch_size, num_channels, pred_len)

        # Head
        if self.target_channels is not None:
            out = out[
                :, self.target_channels, :
            ]  # (batch_size, len(target_channels), pred_len)
            B, C, L = out.size()

            if not self.channel_independent:
                for module in self.head:
                    out = module(out)
            else:
                out = self.head(out)
            out = out.reshape(B, C, -1)  # (batch_size, len(target_channels), pred_len)

        # RevOUT
        if self.revout:
            out = self.revin(out, mode="denorm")

        return out


if __name__ == "__main__":
    # Forecasting
    batch_size = 32
    seq_len = 512
    num_channels = 7
    pred_len = 96
    x = torch.randn(batch_size, num_channels, seq_len)

    model = Linear(
        seq_len=seq_len,
        pred_len=pred_len,
        num_channels=num_channels,
        target_channels=[1, 3],
        revin=True,
        revout=True,
        revin_affine=False,
    )

    output = model(x)



import warnings

warnings.filterwarnings("ignore")

import os
import yaml
import time
from datetime import datetime
from pydantic import BaseModel, Field
from typing import Set, Tuple, List, Optional, Dict, Union

import random
import string


def generate_random_id(length=10):
    # Seed the random number generator with current time and os-specific random data
    random.seed(int(time.time() * 1000) ^ os.getpid())

    characters = string.ascii_letters + string.digits
    return "".join(random.choice(characters) for _ in range(length))


class Experiment(BaseModel):
    model_id: str = Field(
        default="PatchTSTBlind",
        description="Model ID. Options: 'PatchTSTOG', 'PatchTSTBlind', 'JEPA', 'DualJEPA'",
    )
    seed_list: List[int] = Field(
        default=[2024],
        description="List of random seeds to run a single experiment on.",
    )
    seed: int = Field(default=2024, description="Random seed")
    learning_type: str = Field(
        default="sl", description="Type of learning: 'sl', 'ssl'"
    )
    id: str = Field(
        default_factory=generate_random_id,
        description="Experiment ID, randomly generated 10-character string",
    )
    neptune: bool = Field(
        default=False,
        description="Whether to use Neptune for logging. If False, offline logging (JSON) will be used.",
    )
    api_token: str = Field(
        default=os.environ.get("NEPTUNE_API_TOKEN", ""), description="Neptune API token"
    )
    project_name: str = Field(
        default="xmootoo/soz-localization", description="Neptune project name"
    )
    time: str = Field(
        default=datetime.now().strftime("%Y-%m-%d_%H-%M-%S"),
        description="Neptune run ID",
    )
    early_stopping: bool = Field(
        default=True, description="Whether to use early stopping"
    )
    acc: bool = Field(
        default=False, description="Evaluate accuracy or not for classification tasks"
    )
    ch_acc: bool = Field(
        default=False,
        description="Evaluate whole channel accuracy or not for classification tasks",
    )
    thresh: float = Field(
        default=0.5, description="The threshold for binary classification"
    )
    mae: bool = Field(
        default=False, description="Evaluate mean absolute error or not for forecasting"
    )
    other_metrics: bool = Field(
        default=False,
        description="Whether to use other metrics for classification evaluation (e.g., F1-score",
    )
    best_model_metric: str = Field(
        default="loss",
        description="Metric to use for model saving and early stopping. Options: 'loss', 'acc', 'ch_acc'",
    )
    mps: bool = Field(
        default=False,
        description="Whether to use MPS for Apple silicon hardware acceleration",
    )
    grid_search: bool = Field(
        default=False,
        description="Whether to use grid search for hyperparameter tuning",
    )
    task: str = Field(
        default="classification",
        description="Task type. Options: 'forecasting', 'classification'",
    )
    gpu_id: int = Field(
        default=0, description="GPU ID to use for for single device training"
    )
    ablation_id: int = Field(
        default=1, description="Ablation ID for the base experiment"
    )


class Data(BaseModel):
    dtype: str = Field(
        default="float32", description="Type of data. Options: 'float32', 'float64'"
    )
    seq_len: int = Field(default=512, description="Sequence length of the input.")
    window_stride: int = Field(
        default=1, description="Window stride for generating windows."
    )
    pred_len: int = Field(
        default=96, description="Prediction length of the forecast window."
    )
    num_classes: int = Field(
        default=2, description="Number of classes for classification tasks."
    )
    num_channels: int = Field(default=321, description="Number of time series channels")
    drop_last: bool = Field(
        default=False, description="Whether to drop the last batch."
    )
    scale: bool = Field(default=True, description="Normalize data along each channel.")
    balance: bool = Field(
        default=True, description="Balance classes within dataset for classification."
    )
    train_split: float = Field(
        default=0.6, description="Portion of data to use for training"
    )
    val_split: float = Field(
        default=0.2, description="Portion of data used for validation"
    )
    num_workers: int = Field(
        default=4, description="Number of workers for the dataloader"
    )
    pin_memory: bool = Field(
        default=True, description="Whether to pin memory for the dataloader"
    )
    prefetch_factor: int = Field(
        default=2, description="Prefetch factor for the dataloader"
    )
    shuffle_test: bool = Field(
        default=False, description="Whether to shuffle the test set"
    )
    shuffle: bool = Field(default=True, description="Whether to shuffle all datasets")
    patching: bool = Field(
        default=False, description="Whether to use patching for the dataset (LSTM only)"
    )
    patch_dim: int = Field(default=16, description="Patch dimension or patch length.")
    patch_stride: int = Field(
        default=8, description="Patch stride for generating patches.}"
    )
    univariate: bool = Field(
        default=False,
        description="Whether to process a multivariate time series as univariate data (mixed together but separated by channel)",
    )
    seq_load: bool = Field(
        default=False,
        description="Whether to use sequential dataloading. Loads train datasets first, and then test set at test time.",
    )
    numpy_data: bool = Field(
        default=False,
        description="Whether to use numpy data in dataloading (will not be converted to torch tensors).",
    )
    difference_input: bool = Field(
        default=False,
        description="Whether to use 1st-order differencing on the input data for forecasting.",
    )
    scaling: bool = Field(
        default=False,
        description="Whether to use z-score normalize each channel using StandardScaler (scikit-learn).",
    )

    # Option/Underlying Parameters
    root: str = Field(
        default="AAPL", description="The root symbol of the underlying security"
    )
    start_date: str = Field(
        default="20231107",
        description="Start date of the dataset of multiple option contracts",
    )
    end_date: str = Field(
        default="20241114",
        description="End date of the dataset of multiple option contracts",
    )
    interval_min: int = Field(
        default=1,
        description="Interval length in minutes for requested data. Mininimum is 1minute resolution.",
    )
    right: str = Field(
        default="C",
        description="Right of the option of contract. 'C' for call and 'P' for put.",
    )
    moneyness: str = Field(
        default="ATM",
        description="Moneyness of the option contract. Options: 'ATM', 'ITM', or 'OTM'.",
    )
    target_band: float = Field(
        default=0.05,
        description="Target band in proportion between 0 and 1 used for fixed strike selection (e.g., target_band=0.05 uses +/- 5% of current underlying price",
    )
    volatility_scaled: bool = Field(
        default=True,
        description="Whether to select strikes based on scaled historical volatility",
    )
    volatility_scalar: float = Field(
        default=0.1,
        description="Scalar to amplify or attenuate the effect of historical volatility on selecting strikes",
    )
    volatility_type: str = Field(
        default="period",
        description="The time range to calculate historical volatility for volatility-based strike selection",
    )
    target_tte: int = Field(
        default=30,
        description="Desired length of the contract in days. The find_optimal() method of the Contract class will attempt to the contract with the closests TTE",
    )
    tte_tolerance: Tuple[int, int] = Field(
        default=(15, 45),
        description="Tolerance range for selecting a contract if exact target_tte does not exist",
    )
    contract_stride: int = Field(
        default=5,
        description="Stride length used to select multiple contracts at different dates",
    )
    target_channels: Optional[List[str]] = Field(
        default=["option_returns"],
        description="Target channels used in the target window.",
    )
    clean_up: bool = Field(
        default=True,
        description="Whether to clean up the CSV files after saving data from ThetaData API.",
    )
    offline: bool = Field(
        default=False,
        description="Whether to use offline data instead of calling ThetaData API directly.",
    )
    save_dir: str = Field(
        default="../historical_data/", description="Directory to save the data."
    )
    verbose: bool = Field(default=True, description="Verbose print for data loading.")
    intraday: bool = Field(
        default=False,
        description="Whether to use intraday data for the dataset or allow crossover between different days.",
    )

    # Features
    core_feats: List[str] = Field(
        default=["option_mid_price"],
        description="Core features of option and underlying data",
    )
    tte_feats: List[str] = Field(
        default=["sqrt"],
        description="Time-to-Expiration (TTE) features for option contracts.",
    )
    datetime_feats: List[str] = Field(
        default=["sin_minute_of_day"],
        description="Datetime features to use for the time series.",
    )


class Conformal(BaseModel):
    conf: bool = Field(
        default=False,
        description="Whether to use conformal prediction. This will split .",
    )
    alpha: float = Field(
        default=0.1,
        description="Significance level (i.e. error rate) for the conformal prediction.",
    )
    corrected: bool = Field(
        default=False,
        description="Whether to use Bonferonni correction for conformal time series forecasting. This will use the corrected_critical_scores.",
    )
    intervals: bool = Field(
        default=True,
        description="Whether to report interval width mean and standard deviation for evaluation metrics.",
    )
    validation_eval: bool = Field(
        default=False,
        description="Whether evaluate the model for coverage on the validation set at each epoch. If False, it only evaluates on the test set.",
    )


class Train(BaseModel):
    optimizer: str = Field(
        default="adam",
        description="Optimizer for supervised learning: 'adam' or 'adamw'",
    )
    criterion: str = Field(
        default="MSE",
        description="Criterion for supervised learning: 'MSE', 'SmoothL1', 'CrossEntropy', 'BCE', 'ChannelLossBCE', 'ChannelLossCE'",
    )
    num_enc_layers: int = Field(
        default=3, description="Number of encoder layers in the model"
    )
    dropout: float = Field(
        default=0.05, description="Dropout for some of the linears layers in PatchTSTOG"
    )
    batch_first: bool = Field(
        default=True, description="Whether the first dimension is batch"
    )
    norm_mode: str = Field(
        default="batch1d",
        description="Normalization mode: 'batch1d', 'batch2d', or 'layer'",
    )
    batch_size: int = Field(default=64, description="Batch size")
    revin: bool = Field(
        default=True, description="Whether to use instance normalization with RevIN."
    )
    revout: bool = Field(
        default=True, description="Whether to use add mean and std back after forecast."
    )
    revin_affine: bool = Field(
        default=True,
        description="Whether to use learnable affine parameters for RevIN.",
    )
    eps_revin: float = Field(
        default=1e-5, description="Epsilon value for reversible input"
    )
    lr: float = Field(default=1e-4, description="Learning rate")
    epochs: int = Field(default=100, description="Number of epochs to train")
    scheduler: str = Field(
        default=None, description="Scheduler to use for learning rate annealing"
    )
    weight_decay: float = Field(
        default=1e-6, description="Weight decay for the optimizer"
    )
    dataset_class: str = Field(
        default="forecasting",
        description="Task type: 'forecasting', 'forecasting_og', 'classification', 'JEPA', 'DualJEPA'",
    )
    early_stopping: bool = Field(
        default=False, description="Early stopping for supervised learning."
    )
    head_type: str = Field(
        default="linear",
        description="Head type for supervised learning: 'linear' or 'mlp'",
    )  # CyclicalPatchedForecaster (usable)
    return_head: bool = Field(
        default=False, description="Whether to return the head of the model."
    )
    channel_independent: bool = Field(
        default=False,
        description="Whether to use channel independent linear layers for the head.",
    )


class EarlyStopping(BaseModel):
    patience: int = Field(default=10, description="Patience for early stopping.")
    verbose: bool = Field(
        default=True, description="Verbose print for early stopping class."
    )
    delta: float = Field(
        default=0.0, description="Delta additive to the best scores for early stopping."
    )


class Scheduler(BaseModel):
    warmup_steps: int = Field(
        default=15,
        description="Number of warmup epochs for the scheduler for cosine_warmup",
    )
    start_lr: float = Field(
        default=1e-4,
        description="Starting learning rate for the scheduler for warmup, for cosine_warmup",
    )
    ref_lr: float = Field(
        default=1e-3,
        description="End learning rate for the scheduler after warmp, for cosine_warmup",
    )
    final_lr: float = Field(
        default=1e-6,
        description="Final learning rate by the end of the schedule (starting from ref_lr) for cosine_warmup",
    )
    T_max: int = Field(
        default=100,
        description="Maximum number of epochs for the scheduler for CosineAnnealingLR or cosine_warmup",
    )
    last_epoch: int = Field(
        default=-1,
        description="Last epoch for the scheduler or CosineAnnealingLR or cosine_warmup",
    )
    eta_min: float = Field(
        default=1e-6, description="Minimum learning rate for CosineAnnealingLR"
    )
    pct_start: float = Field(
        default=0.3,
        description="Percentage of the cycle (in number of steps) spent increasing the learning rate for OneCycleLR",
    )
    lradj: str = Field(
        default="type3",
        description="Learning rate adjustment type (ontop of scheduling). Options: 'type3', 'TST'",
    )


class PatchTST(BaseModel):
    num_enc_layers: int = Field(
        default=3, description="Number of encoder layers for the PatchTST model."
    )
    d_model: int = Field(
        default=16, description="Model dimension for the PatchTST model."
    )
    d_ff: int = Field(
        default=128, description="FeedForward dimension for the PatchTST model."
    )
    num_heads: int = Field(
        default=4, description="Number of heads for the PatchTST model."
    )
    attn_dropout: float = Field(
        default=0.3,
        description="Dropout rate for attention mechanism in the PatchTST model.",
    )
    ff_dropout: float = Field(
        default=0.3,
        description="Dropout rate for feedforward mechanism in the PatchTST model.",
    )
    pred_dropout: float = Field(
        default=0.0,
        description="Dropout rate for prediction mechanism in the PatchTST model.",
    )
    norm_mode: str = Field(
        default="batch1d", description="Normalization mode for the PatchTST model."
    )


class DLinear(BaseModel):
    moving_avg: int = Field(
        default=25, description="Moving average window for the DLinear model."
    )
    individual: bool = Field(
        default=False,
        description="Whether to use model channels together or separately.",
    )

    # PatchedForecaster
    final_moving_avg: int = Field(
        default=25,
        description="Moving average window for the final forecast in the DLinear model.",
    )


class TSMixer(BaseModel):
    num_enc_layers: int = Field(
        default=2, description="Number of encoder layers for the TSMixer model."
    )
    dropout: float = Field(
        default=0.3, description="Dropout rate for the TSMixer model."
    )


class TimesNet(BaseModel):
    num_enc_layers: int = Field(
        default=2, description="Number of encoder layers for the TimesNet model."
    )
    d_model: int = Field(
        default=16, description="Model dimension for the TimesNet model."
    )
    d_ff: int = Field(
        default=128, description="FeedForward dimension for the TimesNet model."
    )
    num_kernels: int = Field(
        default=6, description="Number of kernels for the TimesNet model."
    )
    c_out: int = Field(
        default=1, description="Output channels for the TimesNet model for forecasting."
    )
    top_k: int = Field(
        default=3,
        description="Top k amplitudes used for the periodic slicing block in TimesNet.",
    )
    dropout: float = Field(
        default=0.3, description="Dropout rate for the TimesNet model."
    )


class ModernTCN(BaseModel):
    num_enc_layers: List[int] = Field(
        default=[2],
        description="Choose from {1, 2, 3} and can make it a list (in str format a,b,c,...) for multistaging with 5 possible stages [a,b,c,d,e] with each element from {1, 2, 3}. For example [1, 1] or [2, 2, 3].",
    )
    d_model: List[int] = Field(
        default=[16],
        description="The model dimension (i.e. Conv1D channel dimension) for each stage. Choose from {32, 64, 128, 256, 512}. Make a list (in str format a,b,c,...)  for multistaging, length equal to number of stages.",
    )
    ffn_ratio: int = Field(
        default=1,
        description="The expansion factor for the feed-forward networks in each block, d_ffn = d_model*ffn_ratio. Choose from {1, 2, 4, 8}",
    )
    dropout: float = Field(
        default=0.0, description="Dropout rate for the ModernTCN model."
    )
    class_dropout: float = Field(
        default=0.0, description="Dropout rate for the classification head."
    )
    large_size: List[int] = Field(
        default=[9],
        description="Size of the large kernel. Choose from {13, 31, 51, 71}. Make a list (in str format a,b,c,...)  for multistaging, length equal to number of stages.",
    )
    small_size: List[int] = Field(
        default=[5],
        description="Size of the small kernel Set to 5 for all experiments. Make a list (in str format a,b,c,...) for multistaging, length equal to number of stages.",
    )
    dw_dims: List[int] = Field(
        default=[256],
        description="Depthwise dimension for each stage. Set to 256 for all stages. Make a list (in str format a,b,c,...) for multistaging, length equal to number of stages.",
    )


class EMForecaster(BaseModel):
    patch_norm: str = Field(
        default="none",
        description="Normalization mode for the PatchedForecaster model. Options: 'none', 'layer'",
    )
    patch_act: str = Field(
        default="gelu",
        description="Activation function for the PatchedForecaster model",
    )
    patch_embed_dim: int = Field(
        default=128, description="Embedding dimension for the PatchedForecaster model"
    )
    pos_enc: str = Field(
        default="none",
        description="Positional encoding for the PatchedForecaster model. Options: 'none', 'learnable'",
    )
    patch_model_id: str = Field(
        default="TSMixer",
        description="Patch model for the EMForecaster class. Options: 'TSMixer', 'DLinear'.",
    )
    d_model: int = Field(
        default=24,
        description="The hidden MLP dimension for TSMixer when using patch_model_id='TSMixer'.",
    )
    dropout: float = Field(
        default=0.3, description="Dropout rate for the EMForecaster model."
    )
    num_enc_layers: int = Field(
        default=2, description="Number of encoder layers for the EMForecaster model."
    )
    moving_avg: int = Field(
        default=25,
        description="Moving average window for the EMForecaster model when patch_model_id='DLinear'.",
    )


class RecurrentModel(BaseModel):
    d_model: int = Field(
        default=16, description="Model dimension for the RecurrentModel."
    )
    bidirectional: bool = Field(
        default=False,
        description="Whether to use bidirectional LSTM (typically for classification).",
    )
    last_state: bool = Field(
        default=True,
        description="Whether to use the last state hidden of the LSTM (typically for classification).",
    )
    avg_state: bool = Field(
        default=False,
        description="Whether to use the average the hidden states of the LSTM.",
    )
    backbone_id: str = Field(
        default="LSTM",
        description="Backbone for the RecurrentModel class. Options: 'LSTM', 'RNN', 'GRU', 'Mamba'.",
    )
    num_enc_layers: int = Field(
        default=1, description="Number of encoder layers for the RecurrentModel."
    )


class Global(BaseModel):
    exp: Experiment = Experiment()
    data: Data = Data()
    train: Train = Train()
    early_stopping: EarlyStopping = EarlyStopping()
    scheduler: Scheduler = Scheduler()
    patchtst: PatchTST = PatchTST()
    dlinear: DLinear = DLinear()
    tsmixer: TSMixer = TSMixer()
    timesnet: TimesNet = TimesNet()
    moderntcn: ModernTCN = ModernTCN()
    conf: Conformal = Conformal()
    emf: EMForecaster = EMForecaster()
    rnn: RecurrentModel = RecurrentModel()


def load_config(file_path: str) -> Global:
    print(f"Received file_path in load_config: {file_path}")
    print(f"Absolute file_path in load_config: {os.path.abspath(file_path)}")
    with open(file_path, "r") as file:
        config_data = yaml.safe_load(file)
    return Global(**config_data)



from pathlib import Path
import itertools
import yaml
from typing import Dict, List, Any, Tuple
from rich.console import Console
import argparse

from optrade.main import main

SCRIPT_DIR = Path(__file__).resolve().parent


def load_ablation_config(file_path: Path) -> Tuple[Dict[str, List[Any]], str]:
    with open(file_path, "r") as file:
        config = yaml.safe_load(file)
    ablation_config = config.get("ablations", {})
    cc_base = config.get("cc", None)
    return ablation_config, cc_base


def generate_ablation_combinations(
    ablation_config: Dict[str, List[Any]],
) -> List[Dict[str, Any]]:
    keys, values = zip(*ablation_config.items())
    return [dict(zip(keys, v)) for v in itertools.product(*values)]


def grid_search(job_name: str) -> None:
    # Adjust the base_path to use the absolute path
    base_path = SCRIPT_DIR.parent / "jobs"
    ablation_path = base_path / job_name / "ablation.yaml"

    # Rich
    console = Console()

    # Load the ablation configuration
    ablation_config, cc_base = load_ablation_config(ablation_path)

    # Generate all combinations of ablations
    ablation_combinations = generate_ablation_combinations(ablation_config)

    for i, ablation in enumerate(ablation_combinations):
        console.log(f"Running ablation {i} (locally): {ablation}")
        main(job_name=job_name, ablation=ablation, ablation_id=i)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "job_name", type=str, default="forecasting/test", help="Name of the ablation"
    )
    args = parser.parse_args()
    grid_search(args.job_name)



from typing import List, Optional, Dict, Any, Tuple, Union
from pydantic import BaseModel, Field, model_validator
import pandas as pd
import numpy as np
import yfinance as yf
import torch
from torch.utils.data import DataLoader
from sklearn.preprocessing import StandardScaler

from optrade.data.thetadata import load_stock_data_eod
from optrade.utils.market_metrics import (
    get_fama_french_factors,
    factor_categorization,
)
from optrade.utils.stock_categories import (
    ThreeFactorLevel,
    FiveFactorLevel,
    SectorType,
    IndustryType,
)
from optrade.data.forecasting import (
    get_contract_datasets,
    get_forecasting_dataset,
    get_forecasting_loaders,
)
from optrade.data.contracts import ContractDataset


class Universe(BaseModel):
    model_config = {"use_enum_values": True, "arbitrary_types_allowed": True}

    # Date range for data retrieval
    start_date: str = Field(
        default=None, description="Start date for data retrieval in YYYYMMDD format"
    )
    end_date: str = Field(
        default=None, description="End date for data retrieval in YYYYMMDD format"
    )

    # Stock collections (only one can be true at a time)
    sp_500: bool = Field(
        default=False,
        description="If True, use S&P 500 stocks as the candidate universe",
    )
    nasdaq_100: bool = Field(
        default=False,
        description="If True, use NASDAQ 100 stocks as the candidate universe",
    )
    dow_jones: bool = Field(
        default=False,
        description="If True, use Dow Jones Industrial Average stocks as the candidate universe",
    )
    russell_2000: bool = Field(
        default=False,
        description="If True, use Russell 2000 stocks as the candidate universe",
    )

    # Candidate roots (used if no collection is selected)
    candidate_roots: List[str] = Field(
        default=None,
        description="Candidate root symbols, to be filtered by other parameters. Used only if no collection (sp_500, nasdaq_100, etc.) is selected",
    )

    # Factor filters
    volatility: Optional[FiveFactorLevel] = Field(
        default=None,
        description="The volatility of the stock. Options: 'very_low' (bottom 20%), 'low' (20-40%), 'medium' (40-60%), 'high' (60-80%), 'very_high' (top 20%)",
    )
    pe_ratio: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The P/E ratio of the stock. Options: 'low' (bottom 33%), 'medium' (middle 33%), 'high' (top 33%)",
    )
    debt_to_equity: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The debt to equity ratio of the stock. Options: 'low' (bottom 33%), 'medium' (middle 33%), 'high' (top 33%)",
    )
    beta: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The beta of the stock. Options: 'low' (bottom 33%), 'medium' (middle 33%), 'high' (top 33%)",
    )
    market_cap: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The market cap of the stock. Options: 'low' (small cap), 'medium' (mid cap), 'high' (large cap)",
    )
    sector: Optional[SectorType] = Field(
        default=None,
        description="The sector of the stock. Options: 'tech', 'healthcare', 'financial', 'consumer_cyclical', 'consumer_defensive', 'industrial', 'energy', 'materials', 'utilities', 'real_estate', 'communication'",
    )
    industry: Optional[IndustryType] = Field(
        default=None,
        description="The industry of the stock. See IndustryType enum for complete list of options matching Yahoo Finance classifications",
    )
    dividend_yield: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The dividend yield of the stock. Options: 'low' (bottom 33%), 'medium' (middle 33%), 'high' (top 33%)",
    )
    earnings_volatility: Optional[ThreeFactorLevel] = Field(
        default=None,
        description="The earnings volatility of the stock. Options: 'low' (bottom 33%), 'medium' (middle 33%), 'high' (top 33%)",
    )

    # Fama French Factors
    market_beta: Optional[str] = Field(
        default=None,
        description="The market beta of the stock. Options: 'high', 'low', 'neutral'",
    )
    size_beta: Optional[str] = Field(
        default=None,
        description="The size beta of the stock. Options: 'small_cap', 'large_cap', 'neutral'",
    )
    value_beta: Optional[str] = Field(
        default=None,
        description="The value beta of the stock. Options: 'value', 'growth', 'neutral'",
    )
    profitability_beta: Optional[str] = Field(
        default=None,
        description="The profitability beta of the stock. Options: 'robust', 'weak', 'neutral'",
    )
    investment_beta: Optional[str] = Field(
        default=None,
        description="The investment beta of the stock. Options: 'conservative', 'aggressive', 'neutral'",
    )
    ff_mode: str = Field(
        default=None,
        description="Mode for the Fama-French model ('3_factor' or '5_factor'). If None, no FF factors will be used for filtering.",
    )

    # Dataset factors
    core_feats: List[str] = Field(
        default=["option_returns"],
        description="Core features to include in the dataset",
    )
    tte_feats: List[str] = Field(
        default=["tte"],
        description="Time to expiration features to include in the dataset",
    )
    datetime_feats: List[str] = Field(
        default=None, description="Datetime features to include in the dataset"
    )
    contract_stride: int = Field(
        default=1, description="Number of contracts to skip between each contract"
    )
    interval_min: int = Field(default=1, description="Interval in minutes for the data")
    right: str = Field(
        default="C", description="Option type: 'C' for call, 'P' for put"
    )
    target_tte: int = Field(default=30, description="Target time to expiration in days")
    tte_tolerance: Tuple[int, int] = Field(
        default=(25, 35), description="Tolerance range for time to expiration"
    )
    moneyness: str = Field(
        default="ATM", description="Moneyness of the option: 'ATM', 'OTM', 'ITM'"
    )
    target_band: float = Field(
        default=0.05, description="Target band for strike selection"
    )
    volatility_type: str = Field(
        default="period",
        description="Type of volatility to use: 'period', 'annualized'",
    )
    volatility_scaled: bool = Field(
        default=False, description="Whether to used volatility-based strike selection"
    )
    volatility_scalar: float = Field(
        default=1.0,
        description="Scalar used for volatility-based strike selection, i.e. number of SDs of the current price",
    )
    train_split: float = Field(default=0.7, description="Train split ratio")
    val_split: float = Field(default=0.15, description="Validation split ratio")
    save_dir: Optional[str] = Field(
        default=None, description="Directory to save the contract datasets"
    )
    verbose: bool = Field(default=False, description="Whether to print verbose output")

    # Attributes used to store information (do not initialize)
    fundamentals: Dict[str, Dict[str, Any]] = Field(
        default_factory=dict, description="Fundamental data for each stock"
    )
    roots: List[str] = Field(
        default_factory=list,
        description="Root symbols of the stocks in the candidate universe",
    )
    contracts: Dict[str, Dict[str, ContractDataset]] = Field(
        default_factory=dict, description="Contract datasets for each stock"
    )

    @model_validator(mode="after")
    def validate_universe_selection(self) -> "Universe":
        """
        Validates that only one index collection is selected at a time and
        ensures candidate_roots is provided if no index is selected.
        """
        # Identify which index collections are selected
        collections = ["sp_500", "nasdaq_100", "dow_jones"]
        selected = [coll for coll in collections if getattr(self, coll, False)]

        # Check if more than one collection is selected
        if len(selected) > 1:
            raise ValueError(
                f"Only one collection can be selected at a time. You selected: {', '.join(selected)}"
            )

        # If no collection is selected, ensure candidate_roots is provided
        if len(selected) == 0 and not self.candidate_roots:
            raise ValueError(
                "Either select a collection (sp_500, nasdaq_100, etc.) or provide candidate_roots"
            )

        return self

    def set_candidate_roots(self) -> None:
        """
        Fetches constituents of a specified index using public data sources and updates candidate_roots.
        """
        if self.sp_500:
            sp_data = pd.read_html(
                "https://en.wikipedia.org/wiki/List_of_S%26P_500_companies"
            )[0]
            self.roots = sp_data["Symbol"].str.replace(".", "-").tolist()
        elif self.nasdaq_100:
            nasdaq_tables = pd.read_html("https://en.wikipedia.org/wiki/Nasdaq-100")
            for i, table in enumerate(nasdaq_tables):
                if "Symbol" in table.columns:
                    self.roots = table["Symbol"].tolist()
                    break
                elif "Ticker" in table.columns:
                    self.roots = table["Ticker"].tolist()
                    break

            # If we couldn't find the right table, raise an error
            if not self.roots:
                raise ValueError(
                    "Could not find NASDAQ-100 constituents table with expected columns"
                )
        elif self.dow_jones:
            # From the output, we can see Table 2 has the right structure
            dj_tables = pd.read_html(
                "https://en.wikipedia.org/wiki/Dow_Jones_Industrial_Average"
            )
            # Table 2 has 'Symbol' column
            self.roots = dj_tables[2]["Symbol"].tolist()
        else:
            self.roots = self.candidate_roots

    def get_fundamentals(self) -> None:
        """
        Retrieves fundamental data for each stock in candidate_roots using yfinance.
        Only includes metrics that are specified in the filter criteria.
        """
        self.fundamentals = dict()

        for root in self.roots:
            fundamental_data = {}
            info = yf.Ticker(root).info

            # Only calculate volatility if the filter is set
            if self.volatility is not None:
                try:
                    volatility = load_stock_data_eod(
                        root=root,
                        start_date=self.start_date,
                        end_date=self.end_date,
                        clean_up=True,
                        offline=False,
                    )["close"].pct_change().std() * (252**0.5)
                    fundamental_data["volatility"] = volatility
                except:
                    volatility = None

            # Add other metrics only if their corresponding filter is set
            factor_map = {
                "pe_ratio": self.pe_ratio is not None and info.get("trailingPE"),
                "debt_to_equity": self.debt_to_equity is not None
                and info.get("debtToEquity"),
                "beta": self.beta is not None and info.get("beta"),
                "market_cap": self.market_cap is not None and info.get("marketCap"),
                "sector": self.sector is not None and info.get("sector").lower(),
                "industry": self.industry is not None and info.get("industry").lower(),
                "dividend_yield": self.dividend_yield is not None
                and info.get("dividendYield") * 100,
            }

            # Add each metric to fundamental_data only if it should be included
            for key, value in factor_map.items():
                if (
                    value
                ):  # This will be false if either the filter is None or the value is None/falsy
                    fundamental_data[key] = value

            # If any values are None (i.e. the stock doesn't have that metric), don't add it to the dictionary
            if all(fundamental_data.values()):
                self.fundamentals[root] = fundamental_data
            else:
                self.roots.remove(root)

    def get_fama_french_factors(self) -> None:
        for root in self.roots:
            try:
                ff_factors = get_fama_french_factors(
                    root=root,
                    start_date=self.start_date,
                    end_date=self.end_date,
                    mode=self.ff_mode,
                )
                ff_factor_categories = factor_categorization(
                    ff_factors, mode=self.ff_mode
                )
                self.fundamentals[root].update(ff_factor_categories)
            except:
                self.roots.remove(root)

    def filter_universe(self) -> None:
        """
        Filters the universe of stocks based on the specified criteria.
        - For ThreeFactorLevel: 'low' (0-33%), 'medium' (33-66%), 'high' (66-100%)
        - For FiveFactorLevel: 'very_low' (0-20%), 'low' (20-40%), 'medium' (40-60%), 'high' (60-80%), 'very_high' (80-100%)
        """

        # Store fundamentals in a df
        fundamentals_df = pd.DataFrame(self.fundamentals).T

        if not self.fundamentals:
            print("No fundamental data available. Run get_fundamentals() first.")
            return

        # Create a copy of roots to filter
        filtered_roots = self.roots.copy()

        # Helper function to get percentiles for a given metric
        def get_percentiles(metric, bins=3):
            values = [
                self.fundamentals[root].get(metric)
                for root in self.roots
                if root in self.fundamentals and metric in self.fundamentals[root]
            ]
            values = [v for v in values if v is not None]

            if not values:
                return []

            if bins == 3:  # Terciles
                return [np.percentile(values, 33.33), np.percentile(values, 66.67)]
            elif bins == 5:  # Quintiles
                return [
                    np.percentile(values, 20),
                    np.percentile(values, 40),
                    np.percentile(values, 60),
                    np.percentile(values, 80),
                ]

        # Filter function for three-level factors
        def filter_three_level(metric, level_value):
            if level_value is None:
                return filtered_roots

            percentiles = get_percentiles(metric, bins=3)
            if not percentiles:
                return filtered_roots

            result = []
            for root in filtered_roots:
                if (
                    root not in self.fundamentals
                    or metric not in self.fundamentals[root]
                ):
                    continue

                value = self.fundamentals[root][metric]
                if level_value == "low" and value <= percentiles[0]:
                    result.append(root)
                elif (
                    level_value == "medium" and percentiles[0] < value <= percentiles[1]
                ):
                    result.append(root)
                elif level_value == "high" and value > percentiles[1]:
                    result.append(root)

            return result

        # Filter function for five-level factors
        def filter_five_level(metric, level_value):
            if level_value is None:
                return filtered_roots

            percentiles = get_percentiles(metric, bins=5)
            if not percentiles:
                return filtered_roots

            result = []
            for root in filtered_roots:
                if (
                    root not in self.fundamentals
                    or metric not in self.fundamentals[root]
                ):
                    continue

                value = self.fundamentals[root][metric]
                if level_value == "very_low" and value <= percentiles[0]:
                    result.append(root)
                elif level_value == "low" and percentiles[0] < value <= percentiles[1]:
                    result.append(root)
                elif (
                    level_value == "medium" and percentiles[1] < value <= percentiles[2]
                ):
                    result.append(root)
                elif level_value == "high" and percentiles[2] < value <= percentiles[3]:
                    result.append(root)
                elif level_value == "very_high" and value > percentiles[3]:
                    result.append(root)

            return result

        # Filter function for categorical factors
        def filter_categorical(metric, category_value):
            if category_value is None:
                return filtered_roots

            result = []
            for root in filtered_roots:
                if (
                    root in self.fundamentals
                    and metric in self.fundamentals[root]
                    and str(self.fundamentals[root][metric]).lower()
                    == str(category_value).lower()
                ):
                    result.append(root)

            return result

        # Apply all filters
        # Five-level factors
        filtered_roots = filter_five_level("volatility", self.volatility)

        # Three-level factors
        filtered_roots = filter_three_level("pe_ratio", self.pe_ratio)
        filtered_roots = filter_three_level("debt_to_equity", self.debt_to_equity)
        filtered_roots = filter_three_level("beta", self.beta)
        filtered_roots = filter_three_level("market_cap", self.market_cap)
        filtered_roots = filter_three_level("dividend_yield", self.dividend_yield)

        # Categorical factors
        filtered_roots = filter_categorical("sector", self.sector)
        filtered_roots = filter_categorical("industry", self.industry)

        # Fama French factors
        if self.ff_mode is not None:
            self.get_fama_french_factors()

            # Define a function for direct FF categorical filtering
            def filter_ff_direct(metric, desired_category):
                if desired_category is None:
                    return filtered_roots

                result = []
                for root in filtered_roots:
                    if self.fundamentals[root][metric] == desired_category:
                        result.append(root)

                return result

            # Apply FF factor filters directly using the specified categories
            filtered_roots = filter_ff_direct("market_beta", self.market_beta)
            filtered_roots = filter_ff_direct("size_beta", self.size_beta)
            filtered_roots = filter_ff_direct("value_beta", self.value_beta)

            if self.ff_mode == "5_factor":
                filtered_roots = filter_ff_direct(
                    "profitability_beta", self.profitability_beta
                )
                filtered_roots = filter_ff_direct(
                    "investment_beta", self.investment_beta
                )

        # Update roots to the filtered list
        self.roots = filtered_roots

    def download(self) -> None:
        # Download all data for the filtered universe
        for root in self.roots:

            # Download contracts
            train_contracts, val_contracts, test_contracts = get_contract_datasets(
                root=root,
                start_date=self.start_date,
                end_date=self.end_date,
                contract_stride=self.contract_stride,
                interval_min=self.interval_min,
                right=self.right,
                target_tte=self.target_tte,
                tte_tolerance=self.tte_tolerance,
                moneyness=self.moneyness,
                target_band=self.target_band,
                volatility_type=self.volatility_type,
                volatility_scaled=self.volatility_scaled,
                volatility_scalar=self.volatility_scalar,
                train_split=self.train_split,
                val_split=self.val_split,
                clean_up=False,  # Set to False to download data
                offline=False,  # Set to False to download data
                save_dir=self.save_dir,
                verbose=self.verbose,
            )

            self.contracts[root] = {
                "train": train_contracts,
                "val": val_contracts,
                "test": test_contracts,
            }

            # Download raw data
            get_forecasting_dataset(
                seq_len=self.seq_len,
                pred_len=self.pred_len,
                contracts=train_contracts,
                tte_tolerance=self.tte_tolerance,
                download_only=True,
            )
            get_forecasting_dataset(
                seq_len=self.seq_len,
                pred_len=self.pred_len,
                contracts=val_contracts,
                tte_tolerance=self.tte_tolerance,
                download_only=True,
            )
            get_forecasting_dataset(
                seq_len=self.seq_len,
                pred_len=self.pred_len,
                contracts=test_contracts,
                tte_tolerance=self.tte_tolerance,
                download_only=True,
            )

    def get_forecasting_loaders(
        self,
        root: str,
        offline: bool = False,
        batch_size: int = 32,
        shuffle: bool = True,
        drop_last: bool = False,
        num_workers: int = 4,
        prefetch_factor: int = 2,
        pin_memory: bool = torch.cuda.is_available(),
    ) -> Union[
        Tuple[DataLoader, DataLoader, DataLoader],
        Tuple[DataLoader, DataLoader, DataLoader, StandardScaler],
    ]:

        loaders = get_forecasting_loaders(
            root=root,
            start_date=self.start_date,
            end_date=self.end_date,
            contract_stride=self.contract_stride,
            interval_min=self.interval_min,
            right=self.right,
            target_tte=self.target_tte,
            tte_tolerance=self.tte_tolerance,
            moneyness=self.moneyness,
            target_band=self.target_band,
            volatility_type=self.volatility_type,
            volatility_scaled=self.volatility_scaled,
            volatility_scalar=self.volatility_scalar,
            train_split=self.train_split,
            val_split=self.val_split,
            core_feats=self.core_feats,
            tte_feats=self.tte_feats,
            datetime_feats=self.datetime_feats,
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last,
            num_workers=num_workers,
            prefetch_factor=prefetch_factor,
            pin_memory=pin_memory,
            clean_up=False,
            offline=offline,
            save_dir=self.save_dir,
            verbose=self.verbose,
            scaling=self.scaling,
            intraday=self.intraday,
            target_channels=self.target_channels,
            seq_len=self.seq_len,
            pred_len=self.pred_len,
            dtype=self.dtype,
        )

        return loaders


if __name__ == "__main__":
    # Create a Universe instance
    save_dir = "temp"
    universe = Universe(
        # sp_500=True,
        dow_jones=True,
        # volatility="medium",
        # pe_ratio="low",
        # debt_to_equity="low",
        # beta="medium",
        # market_cap="low",
        # sector="technology",
        # industry="technology",
        # dividend_yield="low",
        # earnings_volatility="low",
        start_date="20210101",
        end_date="20211001",
        # market_beta="neutral",
        ff_mode="5_factor",
        size_beta="large_cap",
        # volatility="high"
        save_dir=save_dir,
    )

    # Fetch the S&P 500 constituents
    universe.set_candidate_roots()
    universe.get_fundamentals()

    # print(universe.fundamentals["NVDA"],
    # universe.fundamentals["AAPL"],
    # universe.fundamentals["MMM"],
    # universe.fundamentals["CAT"],
    # universe.fundamentals["DIS"]
    # )
    print(f"Universe: {universe.roots}")
    # Filter the universe
    universe.filter_universe()
    print(f"Filtered universe: {universe.roots}")

    universe.roots = ["AAPL"]

    universe.download()
    loaders = universe.get_forecasting_loaders(root="AAPL", offline=True)



import pandas as pd
import numpy as np
import pandas_datareader.data as web
import statsmodels.api as sm
from datetime import datetime
import warnings
from typing import Dict, Any

warnings.filterwarnings("ignore", message="The argument 'date_parser' is deprecated")

from optrade.data.thetadata import load_stock_data_eod
from optrade.data.thetadata import load_stock_data


def get_historical_vol(
    stock_data: pd.DataFrame, volatility_type: str = "period"
) -> float:
    """
    Calculate historical volatility using intraday data from regular trading hours (9:30AM-3:59PM).
    Uses mid prices ((bid+ask)/2) for return calculations to avoid bid-ask bounce. Properly accounts
    for overnight return removal between trading days.

    Args:
        stock_data: DataFrame with datetime column in format "YYYY-MM-DD HH:MM:SS"
                   Must be sorted and contain regular intervals during trading hours
        volatility_type: Type of volatility to calculate. Options: "daily", "period", "annualized".

    Returns:
        Volatility value based on the specified type
    """
    # Convert datetime strings to datetime objects
    datetimes = pd.to_datetime(stock_data["datetime"])

    # Verify data is sorted
    assert (
        datetimes.diff().dropna() > pd.Timedelta(0)
    ).all(), "DataFrame must be sorted by datetime in ascending order"

    # Calculate mid prices
    mid_prices = (stock_data["bid"] + stock_data["ask"]) / 2
    unique_dates = stock_data["datetime"].dt.date.unique()

    # Drop NaN values
    unique_dates = unique_dates[~pd.isna(unique_dates)]
    num_trading_days = len(unique_dates)

    # Get intervals per day by counting observations in first full day
    first_day = stock_data[stock_data["datetime"].dt.date == unique_dates[0]]
    intervals_per_day = len(first_day)
    returns_per_day = intervals_per_day - 1

    print(
        f"Intervals per day are {intervals_per_day} and returns per day are {returns_per_day}"
    )

    # Calculate log returns using mid prices
    log_returns = np.log(mid_prices.values[1:] / mid_prices.values[:-1])

    # Remove overnight returns (last interval of day to first interval of next day)
    valid_return_days = (np.arange(len(log_returns)) + 1) % intervals_per_day != 0

    # Validate the valid_return_days
    n = np.random.randint(0, num_trading_days)
    idx = (returns_per_day) + (intervals_per_day * n)
    assert (
        valid_return_days[idx] == False
    ), f"Did not remove correct overnight returns, expected False at index {idx}, got {valid_return_days[idx]}"

    # Get valid log returns values
    valid_returns = log_returns[valid_return_days]

    # Expected valid returns = (returns per day * number of days)
    expected_valid_returns = returns_per_day * num_trading_days

    assert (
        valid_returns.shape[0] == expected_valid_returns
    ), f"Number of valid returns ({valid_returns.shape[0]}) does not match expected ({expected_valid_returns})"

    # Calculate the standard deviation of interval returns
    interval_vol = np.std(valid_returns)

    # Scale up to daily volatility (sqrt of number of returns per day)
    daily_vol = interval_vol * np.sqrt(returns_per_day)

    # Scale to the number of trading days in the dataset
    period_vol = daily_vol * np.sqrt(num_trading_days)

    if volatility_type == "daily":
        return daily_vol
    elif volatility_type == "period":
        return period_vol
    elif volatility_type == "annualized":
        return period_vol * np.sqrt(252 / num_trading_days)
    else:
        raise ValueError(f"Invalid volatility type: {type}")


def get_train_historical_vol(
    root: str,
    start_date: str,
    end_date: str,
    interval_min: int,
    volatility_window: float,
    volatility_type: str,
) -> pd.DataFrame:
    """
    Get historical volatility for a stock over a given time period.

    Args:
        root: Underlying stock symbol
        start_date: Start date for the total dataset in YYYYMMDD format
        end_date: End date for the total dataset in YYYYMMDD format
        interval_min: Interval in minutes for the underlying stock data
        volatility_window: Proportion of total days to use for historical volatility calculation
        volatility_type: Type of historical volatility to use. Options: "daily", "period", "annualized".

    Returns:
        Historical volatility value based on the specified type.
    """

    # Calculate number of days to use for historical volatility
    total_days = (
        pd.to_datetime(end_date, format="%Y%m%d")
        - pd.to_datetime(start_date, format="%Y%m%d")
    ).days
    num_vol_days = int(volatility_window * total_days)
    vol_end_date = (
        pd.to_datetime(start_date, format="%Y%m%d") + pd.Timedelta(days=num_vol_days)
    ).strftime("%Y%m%d")

    stock_data = load_stock_data(
        root=root,
        start_date=start_date,
        end_date=end_date,
        interval_min=interval_min,
        clean_up=True,
    )

    # Select only the first num_vol_days for calculating volatility
    stock_data = stock_data.loc[stock_data["datetime"] <= vol_end_date]

    # Calculate historical volatility
    return get_historical_vol(stock_data, volatility_type)


def get_fama_french_factors(
    root: str,
    start_date: str,
    end_date: str,
    mode: str = "3_factor",
) -> Dict[str, Any]:
    """
    Calculate Fama-French factor model exposures for a stock over the specified period.

    Args:
        root (str): Root symbol of the underlying security
        start_date (str): Start date in YYYYMMDD format
        end_date (str): End date in YYYYMMDD format
        mode (str): Mode for the Fama-French model ("3_factor" or "5_factor")

    Returns:
        Dictionary containing the factor betas:
        - market_beta: Market excess return sensitivity
        - size_beta: Small Minus Big (SMB) factor exposure
        - value_beta: High Minus Low (HML) book-to-market factor exposure
        - profitability_beta: Robust Minus Weak (RMW) profitability factor (5-factor only)
        - investment_beta: Conservative Minus Aggressive (CMA) investment factor (5-factor only)
        - r_squared: Proportion of return variation explained by the factors
    """

    # Suppress the date_parser deprecation warning
    warnings.filterwarnings(
        "ignore", message="The argument 'date_parser' is deprecated"
    )

    # Shift the start_date by -1 day to get returns for the current day
    ff_start_date = datetime.strptime(start_date, "%Y%m%d")
    ff_end_date = datetime.strptime(end_date, "%Y%m%d")

    # Get stock data
    stock_data = load_stock_data_eod(
        root=root,
        start_date=start_date,
        end_date=end_date,
        clean_up=True,
        offline=False,
    )

    # Calculate daily returns
    stock_data["returns"] = stock_data["close"].pct_change().dropna()
    stock_data["Date"] = stock_data["datetime"].dt.date

    # Drop NaN
    stock_data = stock_data.dropna()

    # Drop all other columns besides Date and returns
    stock_data = stock_data[["Date", "returns"]]

    # Get Fama-French factor data based on mode
    if mode == "3_factor":
        ff_data = web.DataReader(
            "F-F_Research_Data_Factors_daily",
            "famafrench",
            start=ff_start_date,
            end=ff_end_date,
        )[0]
        factor_columns = ["Mkt-RF", "SMB", "HML"]
    elif mode == "5_factor":  # 5_factor
        ff_data = web.DataReader(
            "F-F_Research_Data_5_Factors_2x3_daily",
            "famafrench",
            start=ff_start_date,
            end=ff_end_date,
        )[0]
        factor_columns = ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]
    else:
        raise ValueError(f"Invalid mode: {mode}. Choose '3_factor' or '5_factor'.")

    # Convert percentages to decimals
    ff_data = ff_data / 100

    # Truncate ff_data to the same date range as stock_data["Date"]
    valid_dates = pd.DatetimeIndex(stock_data["Date"])
    ff_data = ff_data.loc[ff_data.index.intersection(valid_dates)]

    # Reset index to make Date a column
    ff_data_reset = ff_data.reset_index()
    ff_data_reset["Date"] = ff_data_reset["Date"].dt.date

    # Merge stock_data with ff_data on Date
    aligned_data = pd.merge(stock_data, ff_data_reset, on="Date", how="inner")

    # Linear regression
    X = aligned_data[factor_columns]
    X = sm.add_constant(X)
    y = aligned_data["returns"] - aligned_data["RF"]  # Excess return

    # Ensure y is 1-dimensional
    if isinstance(y, pd.DataFrame):
        y = y.iloc[:, 0]

    # Run regression
    model = sm.OLS(y, X).fit()

    # Prepare results
    result = {
        "market_beta": model.params.get("Mkt-RF", None),
        "size_beta": model.params.get("SMB", None),
        "value_beta": model.params.get("HML", None),
        "r_squared": model.rsquared,
    }

    # Add additional factors for 5-factor model
    if mode == "5_factor":
        result["profitability_beta"] = model.params.get("RMW", None)
        result["investment_beta"] = model.params.get("CMA", None)

    return result


def factor_categorization(
    factors: Dict[str, float], mode: str = "3_factor"
) -> Dict[str, str]:
    """
    Categorize a stock based on its Fama-French factor exposures.

    Args:
        factors (Dict[str, float]): Dictionary containing factor betas for the stock
        mode (str): Mode for the Fama-French model ("3_factor" or "5_factor")

    Returns:
        Dict[str, str]: Dictionary with categorizations for each factor dimension
    """
    categorization = {}

    # Market beta categorization
    if factors["market_beta"] > 1.1:
        categorization["market_beta"] = "high"
    elif factors["market_beta"] < 0.9:
        categorization["market_beta"] = "low"
    else:
        categorization["market_beta"] = "neutral"

    # Size factor categorization
    if factors["size_beta"] > 0.2:
        categorization["size_beta"] = "small_cap"
    elif factors["size_beta"] < -0.2:
        categorization["size_beta"] = "large_cap"
    else:
        categorization["size_beta"] = "neutral"

    # Value factor categorization
    if factors["value_beta"] > 0.2:
        categorization["value_beta"] = "value"
    elif factors["value_beta"] < -0.2:
        categorization["value_beta"] = "growth"
    else:
        categorization["value_beta"] = "neutral"

    # For 5-factor model, add profitability and investment categorizations
    if mode == "5_factor":
        # Profitability factor categorization
        if factors["profitability_beta"] > 0.2:
            categorization["profitability_beta"] = "robust"
        elif factors["profitability_beta"] < -0.2:
            categorization["profitability_beta"] = "weak"
        else:
            categorization["profitability_beta"] = "neutral"

        # Investment factor categorization
        if factors["investment_beta"] > 0.2:
            categorization["investment_beta"] = "conservative"
        elif factors["investment_beta"] < -0.2:
            categorization["investment_beta"] = "aggressive"
        else:
            categorization["investment_beta"] = "neutral"

    return categorization


if __name__ == "__main__":
    # Choose a well-known stock
    symbol = "AAPL"

    # Set test period (1 year)
    start_date = "20230101"  # YYYYMMDD format
    end_date = "20231231"  # YYYYMMDD format

    print(
        f"Testing Fama-French 5-factor exposures for {symbol} from {start_date} to {end_date}"
    )

    # Calculate Fama-French exposures directly using the symbol
    result = get_fama_french_factors(symbol, start_date, end_date, mode="5_factor")

    # Print results
    print("\nFama-French 5-Factor Exposures:")
    print(
        f"Market Beta: {result['market_beta']:.4f}"
        if result["market_beta"] is not None
        else "Market Beta: None"
    )
    print(
        f"Size Beta (SMB): {result['size_beta']:.4f}"
        if result["size_beta"] is not None
        else "Size Beta: None"
    )
    print(
        f"Value Beta (HML): {result['value_beta']:.4f}"
        if result["value_beta"] is not None
        else "Value Beta: None"
    )
    print(
        f"Profitability Beta (RMW): {result['profitability_beta']:.4f}"
        if result["profitability_beta"] is not None
        else "Profitability Beta: None"
    )
    print(
        f"Investment Beta (CMA): {result['investment_beta']:.4f}"
        if result["investment_beta"] is not None
        else "Investment Beta: None"
    )
    print(
        f"R-squared: {result['r_squared']:.4f}"
        if result["r_squared"] is not None
        else "R-squared: None"
    )

    # Categorize the stock based on its factor exposures
    categorization = factor_categorization(result, mode="5_factor")

    # Add this code at the end of your if __name__ == "__main__": block

    # Categorize the stock based on its factor exposures
    categorization = factor_categorization(result, mode="5_factor")

    # Print the categorization results
    print("\nStock Factor Categorization:")
    print(f"Market: {categorization['market_beta']}")
    print(f"Size: {categorization['size_beta']}")
    print(f"Value: {categorization['value_beta']}")
    print(f"Profitability: {categorization['profitability_beta']}")
    print(f"Investment: {categorization['investment_beta']}")



import torch
from torch import Tensor
import torch.nn as nn
from typing import Optional

from optrade.torch.models.patchtst.backbone import PatchTSTBackbone
from optrade.torch.models.utils.revin import RevIN
from optrade.torch.models.utils.patcher import Patcher
from optrade.torch.models.utils.pos_enc import PositionalEncoding
from optrade.torch.models.utils.weight_init import xavier_init


class PatchTST(nn.Module):
    def __init__(
        self,
        num_enc_layers,
        d_model,
        d_ff,
        num_heads,
        num_channels,
        seq_len,
        pred_len,
        attn_dropout=0.0,
        ff_dropout=0.0,
        pred_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
        revin=True,
        revout=True,
        revin_affine=True,
        eps_revin=1e-5,
        patch_dim=16,
        stride=1,
        return_head=True,
        head_type="linear",
        channel_independent=False,  # Head only
        target_channels: Optional[list] = None,  # Head only
    ) -> None:
        super(PatchTST, self).__init__()

        # Parameters
        self.num_patches = int((seq_len - patch_dim) / stride) + 2
        self.num_channels = num_channels
        self.eps_revin = eps_revin
        self.revin_affine = revin_affine
        self.target_channels = target_channels
        self.revout = revout

        # Initialize layers
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        self.patcher = Patcher(patch_dim, stride)
        self.pos_enc = PositionalEncoding(patch_dim, d_model, self.num_patches)
        self.backbone = PatchTSTBackbone(
            num_enc_layers=num_enc_layers,
            d_model=d_model,
            d_ff=d_ff,
            num_heads=num_heads,
            num_channels=num_channels,
            num_patches=self.num_patches,
            pred_len=pred_len,
            attn_dropout=attn_dropout,
            ff_dropout=ff_dropout,
            pred_dropout=pred_dropout,
            batch_first=batch_first,
            norm_mode=norm_mode,
            return_head=return_head,
            head_type=head_type,
            channel_independent=channel_independent,
            target_channels=target_channels,
        )

        # Weight initialization
        self.apply(xavier_init)

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def forward(self, x, y=None, ch_ids=None):

        # RevIN
        if self._revin:
            x = self.revin(x, mode="norm")

        # Patcher
        x = self.patcher(x)

        # Project + Positional Encoding
        x = self.pos_enc(x)

        # Transformer + Linear Head
        x = self.backbone(x, y, ch_ids)

        # RevOUT
        if self.revout:
            x = self.revin(x, mode="denorm")

        return x


if __name__ == "__main__":
    batch_size = 64
    num_channels = 7
    seq_len = 512
    patch_dim = 16
    patch_stride = 8
    pred_len = 96

    model = PatchTST(
        num_enc_layers=3,
        d_model=128,
        d_ff=512,
        num_heads=4,
        num_channels=num_channels,
        seq_len=seq_len,
        pred_len=pred_len,
        attn_dropout=0.0,
        ff_dropout=0.0,
        pred_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
        revin=True,
        revout=False,
        revin_affine=True,
        eps_revin=1e-5,
        patch_dim=patch_dim,
        stride=patch_stride,
        return_head=True,
        head_type="linear",
        channel_independent=False,
        target_channels=[1, 3, 0],
    )

    x = torch.randn(batch_size, num_channels, seq_len)  # (B, M, L)
    out = model(x)
    print(f"Input shape: {x.shape}, Output shape: {out.shape}")



import pandas as pd
import random
import string
import time
import os


def format_time_dynamic(seconds):
    seconds = int(seconds)
    if seconds < 60:
        return f"{seconds}s"
    minutes, seconds = divmod(seconds, 60)
    if minutes < 60:
        return f"{minutes}min {seconds}s"
    hours, minutes = divmod(minutes, 60)
    if hours < 24:
        return f"{hours}hr {minutes}min {seconds}s"
    days, hours = divmod(hours, 24)
    return f"{days}d {hours}hr {minutes}min {seconds}s"


def fill_open_zeros(group: pd.DataFrame) -> pd.DataFrame:
    """Fills any zero values in the 'mid_price' column with the first non-zero value that occurs
    before 9:35 AM within each group of data."""
    if group.iloc[0]["mid_price"] == 0:
        first_nonzero = group[
            (group["mid_price"] != 0)
            & (group["datetime"].dt.time <= pd.Timestamp("09:35").time())
        ]["mid_price"].iloc[0]
        group.loc[group["mid_price"] == 0, "mid_price"] = first_nonzero
    return group


def generate_random_id(length=10):
    # Seed the random number generator with current time and os-specific random data
    random.seed(int(time.time() * 1000) ^ os.getpid())

    characters = string.ascii_letters + string.digits
    return "".join(random.choice(characters) for _ in range(length))
